import pytest

from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_node import PixelNode
from hipscat.pixel_tree.pixel_node_type import PixelNodeType
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder


def assert_pixel_tree_has_nodes_in_catalog(tree, catalog):
    """assert tree contains the same nodes as the catalog"""
    assert tree.contains((-1, -1))
    for pixel in catalog.get_healpix_pixels():
        assert tree.contains((pixel.order, pixel.pixel))


def test_pixel_tree_small_sky(small_sky_catalog, small_sky_pixels):
    """test pixel tree on small sky"""
    pixel_tree = PixelTreeBuilder.from_healpix(small_sky_catalog.get_healpix_pixels())
    assert len(pixel_tree) == len(small_sky_catalog.get_healpix_pixels()) + 1
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_catalog)
    small_sky_pixel = pixel_tree.get_node(small_sky_pixels[0])
    assert small_sky_pixel.parent == pixel_tree.root_pixel
    assert pixel_tree.root_pixel.node_type == PixelNodeType.ROOT


def test_pixel_tree_small_sky_order1(small_sky_order1_catalog, small_sky_order1_pixels):
    """test pixel tree on small sky order1"""
    pixel_tree = PixelTreeBuilder.from_healpix(small_sky_order1_catalog.get_healpix_pixels())
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_order1_catalog)
    first_pixel = pixel_tree.get_node(small_sky_order1_pixels[0])
    second_pixel = pixel_tree.get_node(small_sky_order1_pixels[1])
    parent_node = first_pixel.parent
    assert parent_node.node_type == PixelNodeType.INNER
    assert first_pixel.parent == second_pixel.parent
    assert parent_node.parent == pixel_tree.root_pixel


def test_pixel_tree_small_sky_from_list(small_sky_catalog, small_sky_pixels):
    """test pixel tree on small sky"""
    pixel_tree = PixelTreeBuilder.from_healpix(small_sky_pixels)
    assert len(pixel_tree) == len(small_sky_pixels) + 1
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_catalog)
    small_sky_pixel = pixel_tree.get_node(small_sky_pixels[0])
    assert small_sky_pixel.parent == pixel_tree.root_pixel
    assert pixel_tree.root_pixel.node_type == PixelNodeType.ROOT


def test_pixel_tree_small_sky_order1_from_list(small_sky_order1_catalog, small_sky_order1_pixels):
    """test pixel tree on small sky order1"""
    pixel_tree = PixelTreeBuilder.from_healpix(small_sky_order1_pixels)
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_order1_catalog)
    first_pixel = pixel_tree.get_node(small_sky_order1_pixels[0])
    second_pixel = pixel_tree.get_node(small_sky_order1_pixels[1])
    parent_node = first_pixel.parent
    assert parent_node.node_type == PixelNodeType.INNER
    assert first_pixel.parent == second_pixel.parent
    assert parent_node.parent == pixel_tree.root_pixel


def test_duplicate_pixel_raises_error():
    """test pixel tree raises error with duplicate pixels"""
    partition_info = [
        HealpixPixel(0, 11),
    ]
    PixelTreeBuilder.from_healpix(partition_info)
    info_with_duplicate = [
        HealpixPixel(0, 11),
        HealpixPixel(0, 11),
    ]
    with pytest.raises(ValueError):
        PixelTreeBuilder.from_healpix(info_with_duplicate)


def test_pixel_duplicated_at_different_order_raises_error():
    """test pixel tree raises error with duplicate pixels at different orders"""
    info_with_duplicate = [
        HealpixPixel(0, 11),
        HealpixPixel(1, 44),  ## overlaps with (0,11)
    ]
    with pytest.raises(ValueError):
        PixelTreeBuilder.from_healpix(info_with_duplicate)


def test_tree_builder_contains_root_node():
    tree_builder = PixelTreeBuilder()
    assert tree_builder.contains(HealpixPixel(-1, -1))
    assert tree_builder.contains((-1, -1))
    assert (-1, -1) in tree_builder
    assert HealpixPixel(-1, -1) in tree_builder
    assert not tree_builder.contains(HealpixPixel(10, 10))
    assert not tree_builder.contains((10, 10))
    assert (10, 10) not in tree_builder
    assert HealpixPixel(10, 10) not in tree_builder


def test_tree_builder_contains_fails_wrong_type():
    tree_builder = PixelTreeBuilder()
    with pytest.raises(TypeError):
        assert tree_builder.contains(-1)


def test_pixel_builder_contains_added_node():
    tree_builder = PixelTreeBuilder()
    assert not tree_builder.contains((0, 0))
    tree_builder.create_node((0, 0), PixelNodeType.LEAF, tree_builder.root_pixel)
    assert tree_builder.contains((0, 0))


def test_pixel_builder_retrieve_added_node():
    tree_builder = PixelTreeBuilder()
    order = 0
    pixel = 0
    node_type = PixelNodeType.LEAF
    tree_builder.create_node((order, pixel), node_type, tree_builder.root_pixel)
    assert isinstance(tree_builder.get_node((order, pixel)), PixelNode)
    assert tree_builder.get_node((order, pixel)).node_type == node_type
    assert tree_builder.get_node((order, pixel)).parent == tree_builder.root_pixel
    assert tree_builder.get_node((order, pixel)).hp_order == order
    assert tree_builder.get_node((order, pixel)).hp_pixel == pixel
    assert tree_builder.get_node(HealpixPixel(order, pixel)) == tree_builder.get_node((order, pixel))
    assert tree_builder[(order, pixel)] == tree_builder.get_node((order, pixel))
    assert tree_builder[HealpixPixel(order, pixel)] == tree_builder.get_node((order, pixel))


def test_create_node_no_parent_node_errors():
    tree_builder = PixelTreeBuilder()
    with pytest.raises(ValueError, match="no parent node exists"):
        tree_builder.create_node((10, 0), PixelNodeType.LEAF)


def test_create_node_leaf_parent_node_errors():
    tree_builder = PixelTreeBuilder()
    tree_builder.create_node((0, 0), PixelNodeType.LEAF)
    with pytest.raises(ValueError, match="has node type leaf"):
        tree_builder.create_node((1, 0), PixelNodeType.LEAF)


def test_create_node_existing_errors():
    tree_builder = PixelTreeBuilder()
    tree_builder.create_node((0, 0), PixelNodeType.INNER)
    with pytest.raises(ValueError, match="node already exists"):
        tree_builder.create_node((0, 0), PixelNodeType.LEAF)


def test_create_node_replace_existing():
    tree_builder = PixelTreeBuilder()
    tree_builder.create_node((0, 0), PixelNodeType.INNER)
    tree_builder.create_node((1, 0), PixelNodeType.LEAF)
    assert tree_builder[(1, 0)].node_type == PixelNodeType.LEAF
    tree_builder.create_node((1, 0), PixelNodeType.INNER, replace_existing_node=True)
    assert tree_builder[(1, 0)].node_type == PixelNodeType.INNER
    node = tree_builder[(0, 0)]
    tree_builder.create_node((0, 0), PixelNodeType.INNER, replace_existing_node=True)
    assert tree_builder[(0, 0)] is not node
    assert tree_builder[(1, 0)].parent is tree_builder[(0, 0)]
    assert tree_builder[(1, 0)] in tree_builder[(0, 0)].children
    tree_builder.create_node((0, 0), PixelNodeType.LEAF, replace_existing_node=True)
    assert (1, 0) not in tree_builder


def test_pixel_builder_remove_added_node():
    tree_builder = PixelTreeBuilder()
    order = 0
    pixel = 0
    node_type = PixelNodeType.LEAF
    tree_builder.create_node((order, pixel), node_type, tree_builder.root_pixel)
    assert (order, pixel) in tree_builder
    assert order in tree_builder.pixels
    assert len(tree_builder.root_pixel.children) == 1
    assert tree_builder.root_pixel.children[0] == tree_builder[(order, pixel)]
    tree_builder.remove_node((order, pixel))
    assert (order, pixel) not in tree_builder
    assert order not in tree_builder.pixels
    assert len(tree_builder.root_pixel.children) == 0


def test_pixel_builder_remove_nodes_descendents():
    tree_builder = PixelTreeBuilder()
    tree_builder.create_node((0, 0), PixelNodeType.INNER, tree_builder.root_pixel)
    tree_builder.create_node((1, 0), PixelNodeType.INNER)
    tree_builder.create_node((2, 0), PixelNodeType.LEAF)
    for order in [0, 1, 2]:
        assert (order, 0) in tree_builder
        assert order in tree_builder.pixels
    assert len(tree_builder.root_pixel.children) == 1
    assert tree_builder.root_pixel.children[0] == tree_builder[(0, 0)]
    tree_builder.remove_node((0, 0))
    for order in [0, 1, 2]:
        assert (order, 0) not in tree_builder
        assert order not in tree_builder.pixels
    assert len(tree_builder.root_pixel.children) == 0


def test_pixel_builder_not_remove_node_order():
    tree_builder = PixelTreeBuilder()
    tree_builder.create_node((0, 0), PixelNodeType.LEAF)
    tree_builder.create_node((0, 1), PixelNodeType.LEAF)
    assert (0, 0) in tree_builder
    assert (0, 1) in tree_builder
    assert 0 in tree_builder.pixels
    assert len(tree_builder.root_pixel.children) == 2
    assert tree_builder.root_pixel.children[0] == tree_builder[(0, 0)]
    assert tree_builder.root_pixel.children[1] == tree_builder[(0, 1)]
    tree_builder.remove_node((0, 0))
    assert (0, 0) not in tree_builder
    assert (0, 1) in tree_builder
    assert 0 in tree_builder.pixels
    assert len(tree_builder.root_pixel.children) == 1
    assert tree_builder.root_pixel.children[0] == tree_builder[(0, 1)]


def test_remove_missing_node_errors():
    tree_builder = PixelTreeBuilder()
    tree_builder.create_node((0, 0), PixelNodeType.LEAF)
    with pytest.raises(ValueError):
        tree_builder.remove_node((1, 1))


def test_pixel_builder_retrieve_none_node():
    tree_builder = PixelTreeBuilder()
    assert tree_builder.get_node((10, 10)) is None
    assert tree_builder.get_node(HealpixPixel(10, 10)) is None
    assert tree_builder[(10, 10)] is None
    assert tree_builder[HealpixPixel(10, 10)] is None


def test_split_leaf_to_match_partitioning():
    builder1 = PixelTreeBuilder()
    builder2 = PixelTreeBuilder()
    builder1.create_node((0, 0), PixelNodeType.LEAF)
    builder2.create_node((0, 0), PixelNodeType.INNER)
    builder2.create_node((1, 0), PixelNodeType.INNER)
    builder2.create_node((2, 0), PixelNodeType.LEAF)
    builder1.split_leaf_to_match_partitioning(builder2[(0, 0)])
    assert builder1[(0, 0)].node_type == PixelNodeType.INNER
    assert builder1[(1, 0)].node_type == PixelNodeType.INNER
    for pixel in [(1, 1), (1, 2), (1, 3)]:
        assert pixel in builder1
        assert builder1[pixel].node_type == PixelNodeType.LEAF
    for pixel in [(2, 0), (2, 1), (2, 2), (2, 3)]:
        assert pixel in builder1
        assert builder1[pixel].node_type == PixelNodeType.LEAF


def test_split_leaf_not_exist_errors():
    builder1 = PixelTreeBuilder()
    builder2 = PixelTreeBuilder()
    builder2.create_node((0, 0), PixelNodeType.INNER)
    builder2.create_node((1, 0), PixelNodeType.INNER)
    builder2.create_node((2, 0), PixelNodeType.LEAF)
    assert (0, 0) not in builder1
    with pytest.raises(ValueError, match="No node in tree"):
        builder1.split_leaf_to_match_partitioning(builder2[(0, 0)])


def test_split_leaf_not_leaf_errors():
    builder1 = PixelTreeBuilder()
    builder2 = PixelTreeBuilder()
    builder1.create_node((0, 0), PixelNodeType.INNER)
    builder2.create_node((0, 0), PixelNodeType.INNER)
    builder2.create_node((1, 0), PixelNodeType.INNER)
    builder2.create_node((2, 0), PixelNodeType.LEAF)
    assert builder1[(0, 0)].node_type == PixelNodeType.INNER
    with pytest.raises(ValueError, match="is not a leaf node"):
        builder1.split_leaf_to_match_partitioning(builder2[(0, 0)])


def test_add_all_descendants_from_node():
    builder1 = PixelTreeBuilder()
    builder2 = PixelTreeBuilder()
    builder1.create_node((0, 0), PixelNodeType.INNER)
    builder2.create_node((0, 0), PixelNodeType.INNER)
    builder2.create_node((1, 0), PixelNodeType.INNER)
    builder2.create_node((1, 1), PixelNodeType.LEAF)
    builder2.create_node((2, 0), PixelNodeType.LEAF)
    builder1.add_all_descendants_from_node(builder2[(0, 0)])
    for pixel in [(1, 0), (1, 1), (2, 0)]:
        assert pixel in builder1
        assert builder1[pixel].node_type == builder2[pixel].node_type


def test_add_all_descendants_from_node_missing_node_errors():
    builder1 = PixelTreeBuilder()
    builder2 = PixelTreeBuilder()
    builder2.create_node((0, 0), PixelNodeType.INNER)
    builder2.create_node((1, 0), PixelNodeType.INNER)
    builder2.create_node((2, 0), PixelNodeType.LEAF)
    assert (0, 0) not in builder1
    with pytest.raises(ValueError, match="No node in tree"):
        builder1.add_all_descendants_from_node(builder2[(0, 0)])


def test_built_pixel_tree_contains_same_nodes_as_builder():
    tree_builder = PixelTreeBuilder()
    tree_builder.create_node((0, 0), PixelNodeType.INNER, tree_builder.root_pixel)
    tree_builder.create_node((1, 1), PixelNodeType.LEAF, tree_builder[(0, 0)])
    tree = tree_builder.build()
    for order, nodes_at_order in tree_builder.pixels.items():
        for pixel, node in nodes_at_order.items():
            assert tree.get_node((order, pixel)) == node
            assert tree_builder.get_node((order, pixel)) == node



from typing import Any, Dict, Union

import pandas as pd
from typing_extensions import Self

from hipscat.io import FilePointer, file_io


class PartitionJoinInfo:
    """Association catalog metadata with which partitions matches occur in the join"""

    PRIMARY_ORDER_COLUMN_NAME = "Norder"
    PRIMARY_PIXEL_COLUMN_NAME = "Npix"
    JOIN_ORDER_COLUMN_NAME = "join_Norder"
    JOIN_PIXEL_COLUMN_NAME = "join_Npix"

    COLUMN_NAMES = [
        PRIMARY_PIXEL_COLUMN_NAME,
        PRIMARY_PIXEL_COLUMN_NAME,
        JOIN_ORDER_COLUMN_NAME,
        JOIN_PIXEL_COLUMN_NAME,
    ]

    def __init__(self, join_info_df: pd.DataFrame) -> None:
        self.data_frame = join_info_df
        self._check_column_names()

    def _check_column_names(self):
        for column in self.COLUMN_NAMES:
            if column not in self.data_frame.columns:
                raise ValueError(f"join_info_df does not contain column {column}")

    @classmethod
    def read_from_file(
        cls, partition_join_info_file: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ) -> Self:
        """Read partition join info from a `partition_join_info.csv` file to create an object

        Args:
            partition_join_info_file: FilePointer to the `partition_join_info.csv` file
            storage_options: dictionary that contains abstract filesystem credentials

        Returns:
            A `PartitionJoinInfo` object with the data from the file
        """
        if not file_io.does_file_or_directory_exist(
            partition_join_info_file, storage_options=storage_options
        ):
            raise FileNotFoundError(
                f"No partition info found where expected: {str(partition_join_info_file)}"
            )

        data_frame = file_io.load_csv_to_pandas(partition_join_info_file, storage_options=storage_options)
        return cls(data_frame)



"""Container class to hold catalog metadata and partition iteration"""
from __future__ import annotations

import dataclasses
from typing import Any, Dict, List, Union

import healpy as hp
import numpy as np
from typing_extensions import TypeAlias

from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.healpix_dataset.healpix_dataset import HealpixDataset, PixelInputTypes
from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_math.cone_filter import filter_pixels_by_cone
from hipscat.pixel_tree.pixel_node_type import PixelNodeType


class Catalog(HealpixDataset):
    """A HiPSCat Catalog with data stored in a HEALPix Hive partitioned structure

    Catalogs of this type are partitioned spatially, contain `partition_info` metadata specifying
    the pixels in Catalog, and on disk conform to the parquet partitioning structure
    `Norder=/Dir=/Npix=.parquet`
    """

    HIPS_CATALOG_TYPES = [CatalogType.OBJECT, CatalogType.SOURCE, CatalogType.MARGIN]

    # Update CatalogInfoClass, used to check if the catalog_info is the correct type, and
    # set the catalog info to the correct type
    CatalogInfoClass: TypeAlias = CatalogInfo
    catalog_info: CatalogInfoClass

    def __init__(
        self,
        catalog_info: CatalogInfoClass,
        pixels: PixelInputTypes,
        catalog_path: str = None,
        storage_options: Union[Dict[Any, Any], None] = None,
    ) -> None:
        """Initializes a Catalog

        Args:
            catalog_info: CatalogInfo object with catalog metadata
            pixels: Specifies the pixels contained in the catalog. Can be either a
                list of HealpixPixel, `PartitionInfo object`, or a `PixelTree` object
            catalog_path: If the catalog is stored on disk, specify the location of the catalog
                Does not load the catalog from this path, only store as metadata
            storage_options: dictionary that contains abstract filesystem credentials
        """
        if catalog_info.catalog_type not in self.HIPS_CATALOG_TYPES:
            raise ValueError(
                f"Catalog info `catalog_type` must be one of "
                f"{', '.join([t.value for t in self.HIPS_CATALOG_TYPES])}"
            )
        super().__init__(catalog_info, pixels, catalog_path, storage_options)

    def filter_by_cone(self, ra: float, dec: float, radius: float) -> Catalog:
        """Filter the pixels in the catalog to only include the pixels that overlap with a cone

        Args:
            ra (float): Right Ascension of the center of the cone in degrees
            dec (float): Declination of the center of the cone in degrees
            radius (float): Radius of the cone in degrees

        Returns:
            A new catalog with only the pixels that overlap with the specified cone
        """
        filtered_cone_pixels = filter_pixels_by_cone(self.pixel_tree, ra, dec, radius)
        filtered_catalog_info = dataclasses.replace(
            self.catalog_info,
            total_rows=None,
        )
        return Catalog(filtered_catalog_info, filtered_cone_pixels)

    # pylint: disable=too-many-locals
    def generate_negative_tree_pixels(self) -> List[HealpixPixel]:
        """Get the leaf nodes at each healpix order that have zero catalog data.

        For example, if an example catalog only had data points in pixel 0 at
        order 0, then this method would return order 0's pixels 1 through 11.
        Used for getting full coverage on margin caches.

        Returns:
            List of HealpixPixels representing the 'negative tree' for the catalog.
        """
        max_depth = self.partition_info.get_highest_order()
        missing_pixels = []
        pixels_at_order = self.pixel_tree.root_pixel.children

        covered_orders = []
        for order_i in range(0, max_depth + 1):
            npix = hp.nside2npix(2**order_i)
            covered_orders.append(np.zeros(npix))

        for order in range(0, max_depth + 1):
            next_order_children = []
            leaf_pixels = []

            for node in pixels_at_order:
                pixel = node.pixel.pixel
                covered_orders[order][pixel] = 1
                if node.node_type == PixelNodeType.LEAF:
                    leaf_pixels.append(pixel)
                else:
                    next_order_children.extend(node.children)

            zero_leafs = np.argwhere(covered_orders[order] == 0).flatten()
            for pix in zero_leafs:
                missing_pixels.append(HealpixPixel(order, pix))
                leaf_pixels.append(pix)

            pixels_at_order = next_order_children

            for order_j in range(order + 1, max_depth + 1):
                explosion_factor = 4 ** (order_j - order)
                for pixel in leaf_pixels:
                    covered_pix = range(pixel * explosion_factor, (pixel + 1) * explosion_factor)
                    covered_orders[order_j][covered_pix] = 1

        return missing_pixels



"""Tests of file IO (reads and writes)"""

import os
import shutil

import numpy.testing as npt
import pandas as pd
import pyarrow as pa

import hipscat.io.write_metadata as io
import hipscat.pixel_math as hist
from hipscat.io import file_io
from hipscat.pixel_math.healpix_pixel import HealpixPixel


def test_write_json_file(assert_text_file_matches, tmp_path):
    """Test of arbitrary json dictionary with strings and numbers"""

    expected_lines = [
        "{\n",
        '    "first_english": "a",',
        '    "first_greek": "alpha",',
        '    "first_number": 1,',
        r'    "first_five_fib": \[',
        "        1,",
        "        1,",
        "        2,",
        "        3,",
        "        5",
        "    ]",
        "}",
    ]

    dictionary = {}
    dictionary["first_english"] = "a"
    dictionary["first_greek"] = "alpha"
    dictionary["first_number"] = 1
    dictionary["first_five_fib"] = [1, 1, 2, 3, 5]

    json_filename = os.path.join(tmp_path, "dictionary.json")
    io.write_json_file(dictionary, json_filename)
    assert_text_file_matches(expected_lines, json_filename)


def test_write_catalog_info(assert_text_file_matches, tmp_path, catalog_info):
    """Test that we accurately write out catalog metadata"""
    catalog_base_dir = os.path.join(tmp_path, "test_name")
    file_io.make_directory(catalog_base_dir)
    expected_lines = [
        "{",
        '    "catalog_name": "test_name",',
        '    "catalog_type": "object",',
        '    "total_rows": 10,',
        '    "epoch": "J2000",',
        '    "ra_column": "ra",',
        '    "dec_column": "dec"',
        "}",
    ]

    io.write_catalog_info(dataset_info=catalog_info, catalog_base_dir=catalog_base_dir)
    metadata_filename = os.path.join(catalog_base_dir, "catalog_info.json")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_write_provenance_info(assert_text_file_matches, tmp_path, catalog_info):
    """Test that we accurately write out tool-provided generation metadata"""
    catalog_base_dir = os.path.join(tmp_path, "test_name")
    file_io.make_directory(catalog_base_dir)
    expected_lines = [
        "{",
        '    "catalog_name": "test_name",',
        '    "catalog_type": "object",',
        '    "total_rows": 10,',
        '    "epoch": "J2000",',
        '    "ra_column": "ra",',
        '    "dec_column": "dec",',
        r'    "version": ".*",',  # version matches digits
        r'    "generation_date": "[.\d]+",',  # date matches date format
        '    "tool_args": {',
        '        "tool_name": "hipscat-import",',
        '        "tool_version": "1.0.0",',
        r'        "input_file_names": \[',
        '            "file1",',
        '            "file2",',
        '            "file3"',
        "        ]",
        "    }",
        "}",
    ]

    tool_args = {
        "tool_name": "hipscat-import",
        "tool_version": "1.0.0",
        "input_file_names": ["file1", "file2", "file3"],
    }

    io.write_provenance_info(
        catalog_base_dir=catalog_base_dir, dataset_info=catalog_info, tool_args=tool_args
    )
    metadata_filename = os.path.join(catalog_base_dir, "provenance_info.json")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_write_partition_info_healpix_pixel_map(assert_text_file_matches, tmp_path):
    """Test that we accurately write out the partition stats for overloaded input"""
    catalog_base_dir = os.path.join(tmp_path, "test_name")
    file_io.make_directory(catalog_base_dir)
    expected_lines = [
        "Norder,Dir,Npix,num_rows",
        "0,0,11,131",
    ]
    pixel_map = {HealpixPixel(0, 11): (131, [11])}
    io.write_partition_info(catalog_base_dir, destination_healpix_pixel_map=pixel_map)
    metadata_filename = os.path.join(catalog_base_dir, "partition_info.csv")
    assert_text_file_matches(expected_lines, metadata_filename)

    expected_lines = [
        "Norder,Dir,Npix,num_rows",
        "1,0,44,51",
        "1,0,45,29",
        "1,0,46,51",
    ]
    pixel_map = {
        HealpixPixel(1, 44): (51, [44]),
        HealpixPixel(1, 45): (29, [45]),
        HealpixPixel(1, 46): (51, [46]),
    }
    io.write_partition_info(catalog_base_dir, destination_healpix_pixel_map=pixel_map)
    metadata_filename = os.path.join(catalog_base_dir, "partition_info.csv")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_write_partition_info_float(assert_text_file_matches, tmp_path):
    """Test that we accurately write out the individual partition stats
    even when the input is floats instead of ints"""
    catalog_base_dir = os.path.join(tmp_path, "test_name")
    file_io.make_directory(catalog_base_dir)
    expected_lines = [
        "Norder,Dir,Npix,num_rows",
        "0,0,11,131",
    ]
    pixel_map = {HealpixPixel(0.0, 11.0): (131, [44.0, 45.0, 46.0])}
    io.write_partition_info(catalog_base_dir, pixel_map)
    metadata_filename = os.path.join(catalog_base_dir, "partition_info.csv")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_write_parquet_metadata(tmp_path, small_sky_dir, basic_catalog_parquet_metadata):
    """Copy existing catalog and create new metadata files for it"""
    catalog_base_dir = os.path.join(tmp_path, "catalog")
    shutil.copytree(
        small_sky_dir,
        catalog_base_dir,
    )
    io.write_parquet_metadata(catalog_base_dir)
    check_parquet_schema(os.path.join(catalog_base_dir, "_metadata"), basic_catalog_parquet_metadata)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        os.path.join(catalog_base_dir, "_common_metadata"),
        basic_catalog_parquet_metadata,
        0,
    )
    ## Re-write - should still have the same properties.
    io.write_parquet_metadata(catalog_base_dir)
    check_parquet_schema(os.path.join(catalog_base_dir, "_metadata"), basic_catalog_parquet_metadata)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        os.path.join(catalog_base_dir, "_common_metadata"),
        basic_catalog_parquet_metadata,
        0,
    )


def test_write_parquet_metadata_order1(tmp_path, small_sky_order1_dir, basic_catalog_parquet_metadata):
    """Copy existing catalog and create new metadata files for it,
    using a catalog with multiple files."""
    temp_path = os.path.join(tmp_path, "catalog")
    shutil.copytree(
        small_sky_order1_dir,
        temp_path,
    )

    io.write_parquet_metadata(temp_path)
    ## 4 row groups for 4 partitioned parquet files
    check_parquet_schema(
        os.path.join(temp_path, "_metadata"),
        basic_catalog_parquet_metadata,
        4,
    )
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        os.path.join(temp_path, "_common_metadata"),
        basic_catalog_parquet_metadata,
        0,
    )


def test_write_index_parquet_metadata(tmp_path):
    """Create an index-like catalog, and test metadata creation."""
    temp_path = os.path.join(tmp_path, "index")

    index_parquet_path = os.path.join(temp_path, "Parts=0", "part_000_of_001.parquet")
    file_io.make_directory(os.path.join(temp_path, "Parts=0"))
    basic_index = pd.DataFrame({"_hipscat_id": [4000, 4001], "ps1_objid": [700, 800]})
    file_io.write_dataframe_to_parquet(basic_index, index_parquet_path)

    index_catalog_parquet_metadata = pa.schema(
        [
            pa.field("_hipscat_id", pa.int64()),
            pa.field("ps1_objid", pa.int64()),
        ]
    )

    io.write_parquet_metadata(temp_path)
    check_parquet_schema(os.path.join(tmp_path, "index", "_metadata"), index_catalog_parquet_metadata)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        os.path.join(tmp_path, "index", "_common_metadata"),
        index_catalog_parquet_metadata,
        0,
    )


def check_parquet_schema(file_name, expected_schema, expected_num_row_groups=1):
    """Check parquet schema against expectations"""
    assert file_io.does_file_or_directory_exist(file_name), f"file not found [{file_name}]"

    single_metadata = file_io.read_parquet_metadata(file_name)
    schema = single_metadata.schema.to_arrow_schema()

    assert len(schema) == len(
        expected_schema
    ), f"object list not the same size ({len(schema)} vs {len(expected_schema)})"

    npt.assert_array_equal(schema.names, expected_schema.names)

    assert schema.equals(expected_schema, check_metadata=False)

    parquet_file = file_io.read_parquet_file(file_name)
    assert parquet_file.metadata.num_row_groups == expected_num_row_groups

    for row_index in range(0, parquet_file.metadata.num_row_groups):
        row_md = parquet_file.metadata.row_group(row_index)
        for column_index in range(0, row_md.num_columns):
            column_metadata = row_md.column(column_index)
            assert column_metadata.file_path.endswith(".parquet")


def test_read_write_fits_point_map(tmp_path):
    """Check that we write and can read a FITS file for spatial distribution."""
    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [51, 29, 51, 0]
    initial_histogram[44:] = filled_pixels[:]
    io.write_fits_map(tmp_path, initial_histogram)

    output_file = os.path.join(tmp_path, "point_map.fits")

    output = file_io.read_fits_image(output_file)
    npt.assert_array_equal(output, initial_histogram)



from __future__ import annotations

import json
import tempfile
from typing import Any, Dict, Union

import healpy as hp
import numpy as np
import pandas as pd
import pyarrow.dataset as pds
import pyarrow.parquet as pq
import yaml

from hipscat.io.file_io.file_pointer import FilePointer, get_fs, strip_leading_slash_for_pyarrow


def make_directory(
    file_pointer: FilePointer, exist_ok: bool = False, storage_options: Union[Dict[Any, Any], None] = None
):
    """Make a directory at a given file pointer

    Will raise an error if a directory already exists, unless `exist_ok` is True in which case
    any existing directories will be left unmodified

    Args:
        file_pointer: location in file system to make directory
        exist_ok: Default False. If false will raise error if directory exists. If true existing
            directories will be ignored and not modified
        storage_options: dictionary that contains abstract filesystem credentials

    Raises:
        OSError
    """
    file_system, file_pointer = get_fs(file_pointer, storage_options=storage_options)
    file_system.makedirs(file_pointer, exist_ok=exist_ok)


def remove_directory(
    file_pointer: FilePointer, ignore_errors=False, storage_options: Union[Dict[Any, Any], None] = None
):
    """Remove a directory, and all contents, recursively.

    Args:
        file_pointer: directory in file system to remove
        ignore_errors: if True errors resulting from failed removals will be ignored
        storage_options: dictionary that contains abstract filesystem credentials
    """

    file_system, file_pointer = get_fs(file_pointer, storage_options)
    if ignore_errors:
        try:
            file_system.rm(file_pointer, recursive=True)
        except Exception:  # pylint: disable=broad-except
            # fsspec doesn't have a "ignore_errors" field in the rm method
            pass
    else:
        file_system.rm(file_pointer, recursive=True)


def write_string_to_file(
    file_pointer: FilePointer,
    string: str,
    encoding: str = "utf-8",
    storage_options: Union[Dict[Any, Any], None] = None,
):
    """Write a string to a text file

    Args:
        file_pointer: file location to write file to
        string: string to write to file
        encoding: Default: 'utf-8', encoding method to write to file with
        storage_options: dictionary that contains abstract filesystem credentials
    """
    file_system, file_pointer = get_fs(file_pointer, storage_options)
    with file_system.open(file_pointer, "w", encoding=encoding) as _file:
        _file.write(string)


def load_text_file(
    file_pointer: FilePointer, encoding: str = "utf-8", storage_options: Union[Dict[Any, Any], None] = None
):
    """Load a text file content to a list of strings.

    Args:
        file_pointer: location of file to read
        encoding: string encoding method used by the file
        storage_options: dictionary that contains abstract filesystem credentials
    Returns:
        text contents of file.
    """
    file_system, file_pointer = get_fs(file_pointer, storage_options)
    with file_system.open(file_pointer, "r", encoding=encoding) as _text_file:
        text_file = _text_file.readlines()

    return text_file


def load_json_file(
    file_pointer: FilePointer, encoding: str = "utf-8", storage_options: Union[Dict[Any, Any], None] = None
) -> dict:
    """Load a json file to a dictionary

    Args:
        file_pointer: location of file to read
        encoding: string encoding method used by the file
        storage_options: dictionary that contains abstract filesystem credentials
    Returns:
        dictionary of key value pairs loaded from the JSON file
    """

    json_dict = None
    file_system, file_pointer = get_fs(file_pointer, storage_options)
    with file_system.open(file_pointer, "r", encoding=encoding) as json_file:
        json_dict = json.load(json_file)

    return json_dict


def load_csv_to_pandas(
    file_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None, **kwargs
) -> pd.DataFrame:
    """Load a csv file to a pandas dataframe

    Args:
        file_pointer: location of csv file to load
        storage_options: dictionary that contains abstract filesystem credentials
        **kwargs: arguments to pass to pandas `read_csv` loading method
    Returns:
        pandas dataframe loaded from CSV
    """
    return pd.read_csv(file_pointer, storage_options=storage_options, **kwargs)


def load_parquet_to_pandas(
    file_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None, **kwargs
) -> pd.DataFrame:
    """Load a parquet file to a pandas dataframe

    Args:
        file_pointer: location of parquet file to load
        storage_options: dictionary that contains abstract filesystem credentials
        **kwargs: arguments to pass to pandas `read_parquet` loading method
    Returns:
        pandas dataframe loaded from parquet
    """
    return pd.read_parquet(file_pointer, storage_options=storage_options, **kwargs)


def write_dataframe_to_csv(
    dataframe: pd.DataFrame,
    file_pointer: FilePointer,
    storage_options: Union[Dict[Any, Any], None] = None,
    **kwargs,
):
    """Write a pandas DataFrame to a CSV file

    Args:
        dataframe: DataFrame to write
        file_pointer: location of file to write to
        storage_options: dictionary that contains abstract filesystem credentials
        **kwargs: args to pass to pandas `to_csv` method
    """
    output = dataframe.to_csv(**kwargs)
    write_string_to_file(file_pointer, output, storage_options=storage_options)


def write_dataframe_to_parquet(
    dataframe: pd.DataFrame, file_pointer, storage_options: Union[Dict[Any, Any], None] = None
):
    """Write a pandas DataFrame to a parquet file

    Args:
        dataframe: DataFrame to write
        file_pointer: location of file to write to
        storage_options: dictionary that contains abstract filesystem credentials
    """
    dataframe.to_parquet(file_pointer, storage_options=storage_options)


def read_parquet_metadata(
    file_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None, **kwargs
) -> pq.FileMetaData:
    """Read FileMetaData from footer of a single Parquet file.

    Args:
        file_pointer: location of file to read metadata from
        storage_options: dictionary that contains abstract filesystem credentials
        **kwargs: additional arguments to be passed to pyarrow.parquet.read_metadata
    """
    file_system, file_pointer = get_fs(file_pointer=file_pointer, storage_options=storage_options)

    file_pointer = strip_leading_slash_for_pyarrow(file_pointer, protocol=file_system.protocol)

    parquet_file = pq.read_metadata(file_pointer, filesystem=file_system, **kwargs)
    return parquet_file


def read_parquet_dataset(dir_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None):
    """Read parquet dataset from directory pointer.

    Args:
        dir_pointer: location of file to read metadata from
        storage_options: dictionary that contains abstract filesystem credentials
    """

    ignore_prefixes = [
        "intermediate",
        "_common_metadata",
        "_metadata",
    ]

    file_system, dir_pointer = get_fs(file_pointer=dir_pointer, storage_options=storage_options)

    # pyarrow.dataset requires the pointer not lead with a slash
    dir_pointer = strip_leading_slash_for_pyarrow(dir_pointer, file_system.protocol)

    dataset = pds.dataset(
        dir_pointer,
        filesystem=file_system,
        exclude_invalid_files=True,
        format="parquet",
        ignore_prefixes=ignore_prefixes,
    )
    return dataset


def read_parquet_file(file_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None):
    """Read parquet file from file pointer.

    Args:
        file_pointer: location of file to read metadata from
        storage_options: dictionary that contains abstract filesystem credentials
    """
    file_system, file_pointer = get_fs(file_pointer, storage_options=storage_options)
    return pq.ParquetFile(file_pointer, filesystem=file_system)


def write_parquet_metadata(
    schema: Any,
    file_pointer: FilePointer,
    metadata_collector: list | None = None,
    storage_options: Union[Dict[Any, Any], None] = None,
    **kwargs,
):
    """Write a metadata only parquet file from a schema

    Args:
        schema: schema to be written
        file_pointer: location of file to be written to
        metadata_collector: where to collect metadata information
        storage_options: dictionary that contains abstract filesystem credentials
        **kwargs: additional arguments to be passed to pyarrow.parquet.write_metadata
    """

    file_system, file_pointer = get_fs(file_pointer=file_pointer, storage_options=storage_options)

    file_pointer = strip_leading_slash_for_pyarrow(file_pointer, protocol=file_system.protocol)
    pq.write_metadata(
        schema, file_pointer, metadata_collector=metadata_collector, filesystem=file_system, **kwargs
    )


def read_fits_image(map_file_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None):
    """Read the object spatial distribution information from a healpix FITS file.

    Args:
        file_pointer: location of file to be written
        storage_options: dictionary that contains abstract filesystem credentials
    Returns:
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
    """
    file_system, map_file_pointer = get_fs(file_pointer=map_file_pointer, storage_options=storage_options)
    with tempfile.NamedTemporaryFile() as _tmp_file:
        with file_system.open(map_file_pointer, "rb") as _map_file:
            map_data = _map_file.read()
            _tmp_file.write(map_data)
            map_fits_image = hp.read_map(_tmp_file.name)
    return map_fits_image


def write_fits_image(
    histogram: np.ndarray, map_file_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
):
    """Write the object spatial distribution information to a healpix FITS file.

    Args:
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        file_pointer: location of file to be written
        storage_options: dictionary that contains abstract filesystem credentials
    """
    file_system, map_file_pointer = get_fs(file_pointer=map_file_pointer, storage_options=storage_options)
    with tempfile.NamedTemporaryFile() as _tmp_file:
        with file_system.open(map_file_pointer, "wb") as _map_file:
            hp.write_map(_tmp_file.name, histogram, overwrite=True, dtype=np.int64)
            _map_file.write(_tmp_file.read())


def read_yaml(file_handle: FilePointer, storage_options: Union[Dict[Any, Any], None] = None):
    """Reads yaml file from filesystem.

    Args:
        file_handle: location of yaml file
        storage_options: dictionary that contains abstract filesystem credentials
    """
    file_system, file_handle = get_fs(file_pointer=file_handle, storage_options=storage_options)
    with file_system.open(file_handle, "r", encoding="utf-8") as _file:
        metadata = yaml.safe_load(_file)
    return metadata


def delete_file(file_handle: FilePointer, storage_options: Union[Dict[Any, Any], None] = None):
    """Deletes file from filesystem.

    Args:
        file_handle: location of file pointer
        storage_options: dictionary that contains filesystem credentials
    """
    file_system, file_handle = get_fs(file_pointer=file_handle, storage_options=storage_options)
    file_system.rm(file_handle)


def read_parquet_file_to_pandas(
    file_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None, **kwargs
) -> pd.DataFrame:
    """Reads a parquet file to a pandas DataFrame

    Args:
        file_pointer (FilePointer): File Pointer to a parquet file
        **kwargs: Additional arguments to pass to pandas read_parquet method

    Returns:
        Pandas DataFrame with the data from the parquet file
    """
    return pd.read_parquet(file_pointer, storage_options=storage_options, **kwargs)



"""Test catalog information is valid"""
import os

from hipscat.catalog import PartitionInfo
from hipscat.io import get_file_pointer_from_path, paths, write_catalog_info
from hipscat.io.validation import is_valid_catalog


def test_is_valid_catalog(tmp_path, small_sky_catalog, small_sky_pixels):
    """Tests if the catalog_info and partition_info files are valid"""
    # An empty directory means an invalid catalog
    catalog_dir_pointer = get_file_pointer_from_path(tmp_path)
    assert not is_valid_catalog(catalog_dir_pointer)

    # Having the catalog_info file is not enough
    write_catalog_info(catalog_dir_pointer, small_sky_catalog.catalog_info)
    assert not is_valid_catalog(catalog_dir_pointer)

    # The catalog is valid if both the catalog_info and partition_info files exist,
    # and the catalog_info is in a valid format
    partition_info_pointer = paths.get_partition_info_pointer(catalog_dir_pointer)
    PartitionInfo.from_healpix(small_sky_pixels).write_to_file(partition_info_pointer)
    assert is_valid_catalog(catalog_dir_pointer)

    # A partition_info file alone is also not enough
    catalog_info_pointer = paths.get_catalog_info_pointer(catalog_dir_pointer)
    os.remove(catalog_info_pointer)
    assert not is_valid_catalog(catalog_dir_pointer)

    # The catalog_info file needs to be in the correct format
    small_sky_catalog.catalog_info.catalog_type = "invalid"
    write_catalog_info(catalog_dir_pointer, small_sky_catalog.catalog_info)
    assert not is_valid_catalog(catalog_dir_pointer)



import os

import numpy as np
import pandas as pd

from hipscat.io.file_io import (
    get_file_pointer_from_path,
    load_csv_to_pandas,
    load_json_file,
    load_parquet_to_pandas,
    load_text_file,
    read_parquet_file_to_pandas,
    write_dataframe_to_csv,
    write_string_to_file,
)
from hipscat.io.paths import pixel_catalog_file


def test_write_string_to_file(tmp_dir_cloud, example_cloud_storage_options):
    test_file_path = os.path.join(tmp_dir_cloud, "text_file.txt")
    test_file_pointer = get_file_pointer_from_path(test_file_path)
    test_string = "this is a test"
    write_string_to_file(
        test_file_pointer, test_string, encoding="utf-8", storage_options=example_cloud_storage_options
    )
    data = load_text_file(test_file_path, encoding="utf-8", storage_options=example_cloud_storage_options)
    assert data[0] == test_string


def test_load_json(small_sky_dir_local, small_sky_dir_cloud, example_cloud_storage_options):
    catalog_cloud_path = os.path.join(small_sky_dir_cloud, "catalog_info.json")
    catalog_info_path = os.path.join(small_sky_dir_local, "catalog_info.json")
    catalog_info_pointer = get_file_pointer_from_path(catalog_info_path)
    json_dict_cloud = load_json_file(catalog_cloud_path, storage_options=example_cloud_storage_options)
    json_dict_local = load_json_file(catalog_info_pointer, encoding="utf-8")
    assert json_dict_cloud == json_dict_local


def test_load_csv_to_pandas(small_sky_dir_local, small_sky_dir_cloud, example_cloud_storage_options):
    partition_info_path = os.path.join(small_sky_dir_local, "partition_info.csv")
    partition_info_path_cloud = os.path.join(small_sky_dir_cloud, "partition_info.csv")
    csv_df = pd.read_csv(partition_info_path)
    loaded_df = load_csv_to_pandas(partition_info_path_cloud, storage_options=example_cloud_storage_options)
    pd.testing.assert_frame_equal(csv_df, loaded_df)


def test_load_parquet_to_pandas(small_sky_dir_local, small_sky_dir_cloud, example_cloud_storage_options):
    pixel_data_path = pixel_catalog_file(small_sky_dir_local, 0, 11)
    pixel_data_path_cloud = pixel_catalog_file(small_sky_dir_cloud, 0, 11)
    parquet_df = pd.read_parquet(pixel_data_path)
    loaded_df = load_parquet_to_pandas(pixel_data_path_cloud, storage_options=example_cloud_storage_options)
    pd.testing.assert_frame_equal(parquet_df, loaded_df)


def test_write_df_to_csv(tmp_dir_cloud, example_cloud_storage_options):
    random_df = pd.DataFrame(np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD"))
    test_file_path = os.path.join(tmp_dir_cloud, "test.csv")
    test_file_pointer = get_file_pointer_from_path(test_file_path)
    write_dataframe_to_csv(
        random_df, test_file_pointer, index=False, storage_options=example_cloud_storage_options
    )
    loaded_df = load_csv_to_pandas(test_file_pointer, storage_options=example_cloud_storage_options)
    pd.testing.assert_frame_equal(loaded_df, random_df)


def test_read_parquet_data(tmp_dir_cloud, example_cloud_storage_options):
    random_df = pd.DataFrame(np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD"))
    test_file_path = os.path.join(tmp_dir_cloud, "test.parquet")
    random_df.to_parquet(test_file_path, storage_options=example_cloud_storage_options)
    file_pointer = get_file_pointer_from_path(test_file_path)
    dataframe = read_parquet_file_to_pandas(file_pointer, storage_options=example_cloud_storage_options)
    pd.testing.assert_frame_equal(dataframe, random_df)



from __future__ import annotations

from dataclasses import dataclass

from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo


@dataclass
class AssociationCatalogInfo(BaseCatalogInfo):
    """Catalog Info for a HiPSCat Association Catalog"""

    primary_catalog: str | None = None
    """Catalog name for the primary (left) side of association"""

    primary_column: str | None = None
    """Column name in the primary (left) side of join"""

    join_catalog: str | None = None
    """Catalog name for the joining (right) side of association"""

    join_column: str | None = None
    """Column name in the joining (right) side of join"""

    required_fields = BaseCatalogInfo.required_fields + [
        "primary_catalog",
        "join_catalog",
    ]

    DEFAULT_TYPE = CatalogType.ASSOCIATION
    REQUIRED_TYPE = CatalogType.ASSOCIATION



import json

import pytest

from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.io import file_io


def test_fields_init(base_catalog_info_data, assert_catalog_info_matches_dict):
    catalog_info = BaseCatalogInfo(**base_catalog_info_data)
    assert_catalog_info_matches_dict(catalog_info, base_catalog_info_data)


def test_required_fields_missing(base_catalog_info_data):
    for field in BaseCatalogInfo.required_fields:
        init_data = base_catalog_info_data.copy()
        init_data[field] = None
        with pytest.raises(ValueError, match=field):
            BaseCatalogInfo(**init_data)


def test_wrong_catalog_type(base_catalog_info_data):
    base_catalog_info_data["catalog_type"] = "wrong_type"
    with pytest.raises(ValueError, match=base_catalog_info_data["catalog_type"]):
        BaseCatalogInfo(**base_catalog_info_data)


def test_str(base_catalog_info_data):
    correct_string = ""
    for name, value in base_catalog_info_data.items():
        correct_string += f"  {name} {value}\n"
    cat_info = BaseCatalogInfo(**base_catalog_info_data)
    assert str(cat_info) == correct_string


def test_read_from_file(base_catalog_info_file, assert_catalog_info_matches_dict):
    base_cat_info_fp = file_io.get_file_pointer_from_path(base_catalog_info_file)
    catalog_info = BaseCatalogInfo.read_from_metadata_file(base_cat_info_fp)
    with open(base_catalog_info_file, "r", encoding="utf-8") as cat_info_file:
        catalog_info_json = json.load(cat_info_file)
        assert_catalog_info_matches_dict(catalog_info, catalog_info_json)



import os

import pytest

from hipscat.io.file_io import (
    append_paths_to_pointer,
    directory_has_contents,
    does_file_or_directory_exist,
    find_files_matching_path,
    get_basename_from_filepointer,
    get_directory_contents,
    get_file_pointer_for_fs,
    get_file_pointer_from_path,
    is_regular_file,
    strip_leading_slash_for_pyarrow,
)
from hipscat.io.file_io.file_pointer import get_fs


def test_get_pointer_from_path(tmp_path):
    tmp_pointer = get_file_pointer_from_path(str(tmp_path))
    assert str(tmp_pointer) == str(tmp_path)


def test_get_basename_from_filepointer(tmp_path):
    catalog_info_string = os.path.join(tmp_path, "catalog_info.json")
    catalog_info_pointer = get_file_pointer_from_path(catalog_info_string)
    assert get_basename_from_filepointer(catalog_info_pointer) == "catalog_info.json"


def test_file_or_dir_exist(small_sky_dir):
    small_sky_pointer = get_file_pointer_from_path(small_sky_dir)
    assert does_file_or_directory_exist(small_sky_pointer)
    catalog_info_string = os.path.join(small_sky_dir, "catalog_info.json")
    catalog_info_pointer = get_file_pointer_from_path(catalog_info_string)
    assert does_file_or_directory_exist(catalog_info_pointer)


def test_file_or_dir_exist_false(small_sky_dir):
    small_sky_pointer = get_file_pointer_from_path(small_sky_dir + "incorrect file")
    assert not does_file_or_directory_exist(small_sky_pointer)


def test_append_paths_to_pointer(tmp_path):
    test_paths = ["folder", "file.txt"]
    test_path = os.path.join(tmp_path, *test_paths)
    tmp_pointer = get_file_pointer_from_path(str(tmp_path))
    assert append_paths_to_pointer(tmp_pointer, *test_paths) == test_path


def test_is_regular_file(small_sky_dir):
    partition_info_file = os.path.join(small_sky_dir, "partition_info.csv")
    assert is_regular_file(partition_info_file)

    assert not is_regular_file(small_sky_dir)

    partition_dir = os.path.join(small_sky_dir, "Norder=0")
    assert not is_regular_file(partition_dir)


def test_find_files_matching_path(small_sky_dir):
    ## no_wildcard
    assert len(find_files_matching_path(small_sky_dir, "partition_info.csv")) == 1

    ## wilcard in the name
    assert len(find_files_matching_path(small_sky_dir, "*.csv")) == 1


def test_find_files_matching_path_directory(small_sky_order1_dir):
    assert len(find_files_matching_path(small_sky_order1_dir)) == 1

    ## wildcard in directory - will match all files at indicated depth
    assert len(find_files_matching_path(small_sky_order1_dir, "*", "*", "*")) == 4


def test_directory_has_contents(small_sky_order1_dir, tmp_path):
    assert directory_has_contents(small_sky_order1_dir)
    assert not directory_has_contents(tmp_path)


def test_get_directory_contents(small_sky_order1_dir, tmp_path):
    small_sky_contents = get_directory_contents(small_sky_order1_dir)

    for i, content in enumerate(small_sky_contents):
        if not content.startswith("/"):
            small_sky_contents[i] = f"/{content}"

    assert len(small_sky_contents) == 4

    expected = [
        os.path.join(small_sky_order1_dir, "Norder=1"),
        os.path.join(small_sky_order1_dir, "catalog_info.json"),
        os.path.join(small_sky_order1_dir, "partition_info.csv"),
        os.path.join(small_sky_order1_dir, "point_map.fits"),
    ]

    assert small_sky_contents == expected

    assert len(get_directory_contents(tmp_path)) == 0


def test_get_fs():
    filesystem, _ = get_fs("file://")
    assert filesystem.protocol == "file"

    # this will fail if the environment installs lakefs to import
    with pytest.raises(ImportError):
        get_fs("lakefs://")

    with pytest.raises(ValueError):
        get_fs("invalid://")


def test_get_file_pointer_for_fs():
    test_abfs_protocol_path = get_file_pointer_from_path("abfs:///container/path/to/parquet/file")
    assert (
        get_file_pointer_for_fs("abfs", file_pointer=test_abfs_protocol_path)
        == "/container/path/to/parquet/file"
    )
    test_s3_protocol_path = get_file_pointer_from_path("s3:///bucket/path/to/catalog.json")
    assert get_file_pointer_for_fs("s3", file_pointer=test_s3_protocol_path) == "/bucket/path/to/catalog.json"
    test_local_path = get_file_pointer_from_path("/path/to/file")
    assert get_file_pointer_for_fs("file", file_pointer=test_local_path) == test_local_path
    test_local_protocol_path = get_file_pointer_from_path("file:///path/to/file")
    assert get_file_pointer_for_fs("file", file_pointer=test_local_protocol_path) == "/path/to/file"


def test_strip_leading_slash_for_pyarrow():
    test_leading_slash_filename = get_file_pointer_from_path("/bucket/path/test.txt")
    assert (
        strip_leading_slash_for_pyarrow(test_leading_slash_filename, protocol="abfs")
        == "bucket/path/test.txt"
    )
    test_non_leading_slash_filenaem = get_file_pointer_from_path("bucket/path/test.txt")
    assert (
        strip_leading_slash_for_pyarrow(test_non_leading_slash_filenaem, protocol="abfs")
        == "bucket/path/test.txt"
    )



import dataclasses
import json

import pytest

from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.margin_cache.margin_cache_catalog_info import MarginCacheCatalogInfo
from hipscat.io import file_io


def test_margin_cache_catalog_info(margin_cache_catalog_info, assert_catalog_info_matches_dict):
    info = MarginCacheCatalogInfo(**margin_cache_catalog_info)
    assert_catalog_info_matches_dict(info, margin_cache_catalog_info)


def test_str(margin_cache_catalog_info):
    correct_string = ""
    for name, value in margin_cache_catalog_info.items():
        correct_string += f"  {name} {value}\n"
    cat_info = MarginCacheCatalogInfo(**margin_cache_catalog_info)
    assert str(cat_info) == correct_string


def test_read_from_file(margin_cache_catalog_info_file, assert_catalog_info_matches_dict):
    cat_info_fp = file_io.get_file_pointer_from_path(margin_cache_catalog_info_file)
    catalog_info = MarginCacheCatalogInfo.read_from_metadata_file(cat_info_fp)
    for column in [
        "catalog_name",
        "catalog_type",
        "total_rows",
        "primary_catalog",
        "margin_threshold",
    ]:
        assert column in dataclasses.asdict(catalog_info)

    with open(margin_cache_catalog_info_file, "r", encoding="utf-8") as cat_info_file:
        catalog_info_json = json.load(cat_info_file)
        assert_catalog_info_matches_dict(catalog_info, catalog_info_json)


def test_required_fields_missing(margin_cache_catalog_info):
    required_fields = ["primary_catalog", "margin_threshold"]
    for required_field in required_fields:
        assert required_field in MarginCacheCatalogInfo.required_fields
    for field in required_fields:
        init_data = margin_cache_catalog_info.copy()
        init_data[field] = None
        with pytest.raises(ValueError, match=field):
            MarginCacheCatalogInfo(**init_data)


def test_type_missing(margin_cache_catalog_info):
    init_data = margin_cache_catalog_info.copy()
    init_data["catalog_type"] = None
    catalog_info = MarginCacheCatalogInfo(**init_data)
    assert catalog_info.catalog_type == CatalogType.MARGIN


def test_wrong_type(margin_cache_catalog_info, catalog_info_data):
    with pytest.raises(TypeError, match="unexpected"):
        MarginCacheCatalogInfo(**catalog_info_data)

    with pytest.raises(ValueError, match="type margin"):
        init_data = margin_cache_catalog_info.copy()
        init_data["catalog_type"] = CatalogType.OBJECT
        MarginCacheCatalogInfo(**init_data)



import dataclasses
import os
import os.path
from typing import List

import pandas as pd
import pytest

from hipscat.catalog import PartitionInfo
from hipscat.catalog.association_catalog.association_catalog_info import AssociationCatalogInfo
from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.inspection.almanac import Almanac
from hipscat.pixel_math import HealpixPixel

DATA_DIR_NAME = "data"
ALMANAC_DIR_NAME = "almanac"
SMALL_SKY_DIR_NAME = "small_sky"
SMALL_SKY_ORDER1_DIR_NAME = "small_sky_order1"
SMALL_SKY_TO_SMALL_SKY_ORDER1_DIR_NAME = "small_sky_to_small_sky_order1"
TEST_DIR = os.path.dirname(__file__)

# pylint: disable=missing-function-docstring, redefined-outer-name


def pytest_addoption(parser):
    parser.addoption("--cloud", action="store", default="abfs")


def pytest_generate_tests(metafunc):
    # This is called for every test. Only get/set command line arguments
    # if the argument is specified in the list of test "fixturenames".
    option_value = metafunc.config.option.cloud
    if "cloud" in metafunc.fixturenames and option_value is not None:
        metafunc.parametrize("cloud", [option_value])


@pytest.fixture
def example_cloud_path(cloud):
    if cloud == "abfs":
        return "abfs:///hipscat/pytests/hipscat"

    else:
        raise NotImplementedError("Cloud format not implemented for hipscat tests!")


@pytest.fixture
def example_cloud_storage_options(cloud):
    if cloud == "abfs":
        storage_options = {
            "account_key": os.environ.get("ABFS_LINCCDATA_ACCOUNT_KEY"),
            "account_name": os.environ.get("ABFS_LINCCDATA_ACCOUNT_NAME"),
        }
        return storage_options

    return {}


@pytest.fixture
def tmp_dir_cloud(example_cloud_path):
    return os.path.join(example_cloud_path, "tmp")


@pytest.fixture
def test_data_dir_cloud(example_cloud_path):
    return os.path.join(example_cloud_path, DATA_DIR_NAME)


@pytest.fixture
def almanac_dir_cloud(test_data_dir_cloud):
    return os.path.join(test_data_dir_cloud, ALMANAC_DIR_NAME)


@pytest.fixture
def small_sky_dir_cloud(test_data_dir_cloud):
    return os.path.join(test_data_dir_cloud, SMALL_SKY_DIR_NAME)


@pytest.fixture
def small_sky_order1_dir_cloud(test_data_dir_cloud):
    return os.path.join(test_data_dir_cloud, SMALL_SKY_ORDER1_DIR_NAME)


@pytest.fixture
def small_sky_to_small_sky_order1_dir_cloud(test_data_dir_cloud):
    return os.path.join(test_data_dir_cloud, SMALL_SKY_TO_SMALL_SKY_ORDER1_DIR_NAME)


@pytest.fixture
def catalog_pixels() -> List[HealpixPixel]:
    return [HealpixPixel(1, 0), HealpixPixel(1, 1), HealpixPixel(2, 8)]


@pytest.fixture
def association_catalog_path_cloud(test_data_dir_cloud) -> str:
    return os.path.join(test_data_dir_cloud, "small_sky_to_small_sky_order1")


@pytest.fixture
def association_catalog_info_file_cloud(association_catalog_path_cloud) -> str:
    return os.path.join(association_catalog_path_cloud, "catalog_info.json")


@pytest.fixture
def index_catalog_info_file_cloud(test_data_dir_cloud) -> str:
    return os.path.join(test_data_dir_cloud, "index_catalog", "catalog_info.json")


@pytest.fixture
def margin_cache_catalog_info_file_cloud(test_data_dir_cloud) -> str:
    return os.path.join(test_data_dir_cloud, "margin_cache", "catalog_info.json")


@pytest.fixture
def source_catalog_info_file_cloud(test_data_dir_cloud) -> str:
    return os.path.join(test_data_dir_cloud, "small_sky_source", "catalog_info.json")


@pytest.fixture
def association_catalog_info(association_catalog_info_data) -> AssociationCatalogInfo:
    return AssociationCatalogInfo(**association_catalog_info_data)


@pytest.fixture
def association_catalog_partition_join_file_cloud(association_catalog_path_cloud) -> str:
    return os.path.join(association_catalog_path_cloud, "partition_join_info.csv")


@pytest.fixture
def dataset_path_cloud(test_data_dir_cloud) -> str:
    return os.path.join(test_data_dir_cloud, "dataset")


@pytest.fixture
def base_catalog_info_file_cloud(dataset_path_cloud) -> str:
    return os.path.join(dataset_path_cloud, "catalog_info.json")


@pytest.fixture
def base_catalog_info(base_catalog_info_data) -> BaseCatalogInfo:
    return BaseCatalogInfo(**base_catalog_info_data)


@pytest.fixture
def catalog_path_cloud(test_data_dir_cloud) -> str:
    return os.path.join(test_data_dir_cloud, "catalog")


@pytest.fixture
def catalog_info_file_cloud(catalog_path_cloud) -> str:
    return os.path.join(catalog_path_cloud, "catalog_info.json")


@pytest.fixture
def test_data_dir():
    return os.path.join(TEST_DIR, DATA_DIR_NAME)


@pytest.fixture
def small_sky_dir_local(test_data_dir):
    return os.path.join(test_data_dir, SMALL_SKY_DIR_NAME)


@pytest.fixture
def small_sky_order1_dir_local(test_data_dir):
    return os.path.join(test_data_dir, SMALL_SKY_ORDER1_DIR_NAME)


@pytest.fixture
def assert_catalog_info_matches_dict():
    def assert_match(catalog_info: BaseCatalogInfo, dictionary: dict):
        """Check that all members of the catalog_info object match dictionary
        elements, where specified."""
        catalog_info_dict = dataclasses.asdict(catalog_info)
        for key, value in dictionary.items():
            assert catalog_info_dict[key] == value

    return assert_match


@pytest.fixture
def base_catalog_info_data() -> dict:
    return {
        "catalog_name": "test_name",
        "catalog_type": "object",
        "total_rows": 10,
    }


@pytest.fixture
def catalog_info_data() -> dict:
    return {
        "catalog_name": "test_name",
        "catalog_type": "object",
        "total_rows": 10,
        "epoch": "J2000",
        "ra_column": "ra",
        "dec_column": "dec",
    }


@pytest.fixture
def association_catalog_info_data() -> dict:
    return {
        "catalog_name": "test_name",
        "catalog_type": "association",
        "total_rows": 10,
        "primary_catalog": "small_sky",
        "primary_column": "id",
        "join_catalog": "small_sky_order1",
        "join_column": "id",
    }


@pytest.fixture
def source_catalog_info() -> dict:
    return {
        "catalog_name": "test_source",
        "catalog_type": "source",
        "total_rows": 100,
        "epoch": "J2000",
        "ra_column": "source_ra",
        "dec_column": "source_dec",
    }


@pytest.fixture
def source_catalog_info_with_extra() -> dict:
    return {
        "catalog_name": "test_source",
        "catalog_type": "source",
        "total_rows": 100,
        "epoch": "J2000",
        "ra_column": "source_ra",
        "dec_column": "source_dec",
        "primary_catalog": "test_name",
        "mjd_column": "mjd",
        "band_column": "band",
        "mag_column": "mag",
        "mag_err_column": "",
    }


@pytest.fixture
def margin_cache_catalog_info() -> dict:
    return {
        "catalog_name": "test_margin",
        "catalog_type": "margin",
        "total_rows": 100,
        "primary_catalog": "test_name",
        "margin_threshold": 0.5,
    }


@pytest.fixture
def index_catalog_info() -> dict:
    return {
        "catalog_name": "test_index",
        "catalog_type": "index",
        "total_rows": 100,
        "primary_catalog": "test_name",
        "indexing_column": "id",
    }


@pytest.fixture
def index_catalog_info_with_extra() -> dict:
    return {
        "catalog_name": "test_index",
        "catalog_type": "index",
        "total_rows": 100,
        "primary_catalog": "test_name",
        "indexing_column": "id",
        "extra_columns": ["foo", "bar"],
    }


@pytest.fixture
def catalog_info(catalog_info_data) -> CatalogInfo:
    return CatalogInfo(**catalog_info_data)


@pytest.fixture
def association_catalog_join_pixels() -> pd.DataFrame:
    return pd.DataFrame.from_dict(
        {
            PartitionJoinInfo.PRIMARY_ORDER_COLUMN_NAME: [0, 0, 0, 0],
            PartitionJoinInfo.PRIMARY_PIXEL_COLUMN_NAME: [11, 11, 11, 11],
            PartitionJoinInfo.JOIN_ORDER_COLUMN_NAME: [1, 1, 1, 1],
            PartitionJoinInfo.JOIN_PIXEL_COLUMN_NAME: [44, 45, 46, 47],
        }
    )


@pytest.fixture
def default_almanac_cloud(example_cloud_path, example_cloud_storage_options):
    """Set up default environment variables and fetch default almanac data."""

    test_data_dir = os.path.join(example_cloud_path, "data")
    almanac_dir = os.path.join(example_cloud_path, "data", "almanac")

    os.environ["HIPSCAT_ALMANAC_DIR"] = almanac_dir
    os.environ["HIPSCAT_DEFAULT_DIR"] = test_data_dir

    return Almanac(storage_options=example_cloud_storage_options)



from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_node import PixelNode
from hipscat.pixel_tree.pixel_node_type import PixelNodeType
from hipscat.pixel_tree.pixel_tree import PixelTree


def test_pixel_tree_length():
    lengths = [1, 3, 6]
    for length in lengths:
        root_node = PixelNode((-1, -1), PixelNodeType.ROOT, None)
        last_node = root_node
        nodes = {-1: {-1: root_node}}
        for order in range(length):
            last_node = PixelNode((order, 0), PixelNodeType.INNER, last_node)
            nodes[order] = {0: last_node}
        tree = PixelTree(root_pixel=root_node, pixels=nodes)
        assert len(tree) == length + 1


def test_pixel_tree_contains():
    root_node = PixelNode((-1, -1), PixelNodeType.ROOT, None)
    leaf_node = PixelNode((0, 0), PixelNodeType.LEAF, root_node)
    nodes = {-1: {-1: root_node}, 0: {0: leaf_node}}
    tree = PixelTree(root_pixel=root_node, pixels=nodes)
    assert tree.contains((0, 0))
    assert tree.contains(HealpixPixel(0, 0))
    assert (0, 0) in tree
    assert HealpixPixel(0, 0) in tree
    assert not tree.contains((1, 1))
    assert not tree.contains(HealpixPixel(1, 1))
    assert (1, 1) not in tree
    assert HealpixPixel(1, 1) not in tree


def test_pixel_tree_get_node():
    root_node = PixelNode((-1, -1), PixelNodeType.ROOT, None)
    leaf_node = PixelNode((0, 0), PixelNodeType.LEAF, root_node)
    nodes = {-1: {-1: root_node}, 0: {0: leaf_node}}
    tree = PixelTree(root_pixel=root_node, pixels=nodes)
    assert tree.get_node((0, 0)) == leaf_node
    assert tree.get_node(HealpixPixel(0, 0)) == leaf_node
    assert tree[(0, 0)] == leaf_node
    assert tree[HealpixPixel(0, 0)] == leaf_node
    assert tree.get_node((1, 1)) is None
    assert tree.get_node(HealpixPixel(1, 1)) is None
    assert tree[(1, 1)] is None
    assert tree[HealpixPixel(1, 1)] is None



"""Utility functions for writing metadata files"""

import dataclasses
import json
from datetime import datetime
from importlib.metadata import version
from typing import Any, Dict, Union

import numpy as np
import pandas as pd

from hipscat.io import file_io, paths


def write_json_file(
    metadata_dictionary: dict,
    file_pointer: file_io.FilePointer,
    storage_options: Union[Dict[Any, Any], None] = None,
):
    """Convert metadata_dictionary to a json string and print to file.

    Args:
        metadata_dictionary (:obj:`dictionary`): a dictionary of key-value pairs
        file_pointer (str): destination for the json file
        storage_options: dictionary that contains abstract filesystem credentials
    """
    dumped_metadata = json.dumps(metadata_dictionary, indent=4)
    file_io.write_string_to_file(file_pointer, dumped_metadata + "\n", storage_options=storage_options)


def write_catalog_info(catalog_base_dir, dataset_info, storage_options: Union[Dict[Any, Any], None] = None):
    """Write a catalog_info.json file with catalog metadata

    Args:
        catalog_base_dir (str): base directory for catalog, where file will be written
        dataset_info (:obj:`BaseCatalogInfo`) base metadata for the catalog
        storage_options: dictionary that contains abstract filesystem credentials
    """
    metadata = dataclasses.asdict(dataset_info)
    catalog_info_pointer = paths.get_catalog_info_pointer(catalog_base_dir)

    write_json_file(metadata, catalog_info_pointer, storage_options=storage_options)


def write_provenance_info(
    catalog_base_dir: file_io.FilePointer,
    dataset_info,
    tool_args: dict,
    storage_options: Union[Dict[Any, Any], None] = None,
):
    """Write a provenance_info.json file with all assorted catalog creation metadata

    Args:
        catalog_base_dir (str): base directory for catalog, where file will be written
        dataset_info (:obj:`BaseCatalogInfo`) base metadata for the catalog
        tool_args (:obj:`dict`): dictionary of additional arguments provided by the tool creating
            this catalog.
        storage_options: dictionary that contains abstract filesystem credentials
    """
    metadata = dataclasses.asdict(dataset_info)
    metadata["version"] = version("hipscat")
    now = datetime.now()
    metadata["generation_date"] = now.strftime("%Y.%m.%d")

    metadata["tool_args"] = tool_args

    metadata_pointer = paths.get_provenance_pointer(catalog_base_dir)
    write_json_file(metadata, metadata_pointer, storage_options=storage_options)


def write_partition_info(
    catalog_base_dir: file_io.FilePointer,
    destination_healpix_pixel_map: dict,
    storage_options: Union[Dict[Any, Any], None] = None,
):
    """Write all partition data to CSV file.

    Args:
        catalog_base_dir (str): base directory for catalog, where file will be written
        destination_healpix_pixel_map (dict):  dictionary that maps the HealpixPixel to a
            tuple of origin pixel information:
            - 0 - the total number of rows found in this destination pixel
            - 1 - the set of indexes in histogram for the pixels at the original healpix order
        storage_options: dictionary that contains abstract filesystem credentials
    """
    partition_info_pointer = paths.get_partition_info_pointer(catalog_base_dir)
    data_frame = pd.DataFrame(destination_healpix_pixel_map.keys())
    # Set column names.
    data_frame.columns = [
        "Norder",
        "Npix",
    ]
    data_frame["num_rows"] = [pixel_info[0] for pixel_info in destination_healpix_pixel_map.values()]
    data_frame["Dir"] = [int(x / 10_000) * 10_000 for x in data_frame["Npix"]]

    # Reorder the columns to match full path, and force to integer types.
    data_frame = data_frame[
        [
            "Norder",
            "Dir",
            "Npix",
            "num_rows",
        ]
    ].astype(int)

    file_io.write_dataframe_to_csv(
        data_frame, partition_info_pointer, index=False, storage_options=storage_options
    )


def write_parquet_metadata(catalog_path, storage_options: Union[Dict[Any, Any], None] = None):
    """Generate parquet metadata, using the already-partitioned parquet files
    for this catalog

    Args:
        catalog_path (str): base path for the catalog
        storage_options: dictionary that contains abstract filesystem credentials
    """

    dataset = file_io.read_parquet_dataset(catalog_path, storage_options=storage_options)
    metadata_collector = []

    for hips_file in dataset.files:
        hips_file_pointer = file_io.get_file_pointer_from_path(hips_file, include_protocol=catalog_path)
        single_metadata = file_io.read_parquet_metadata(hips_file_pointer, storage_options=storage_options)
        relative_path = hips_file[len(catalog_path) :]
        single_metadata.set_file_path(relative_path)
        metadata_collector.append(single_metadata)

    ## Write out the two metadata files
    catalog_base_dir = file_io.get_file_pointer_from_path(catalog_path)
    metadata_file_pointer = paths.get_parquet_metadata_pointer(catalog_base_dir)
    common_metadata_file_pointer = paths.get_common_metadata_pointer(catalog_base_dir)

    file_io.write_parquet_metadata(
        dataset.schema,
        metadata_file_pointer,
        metadata_collector=metadata_collector,
        storage_options=storage_options,
    )
    file_io.write_parquet_metadata(
        dataset.schema, common_metadata_file_pointer, storage_options=storage_options
    )


def write_fits_map(catalog_path, histogram: np.ndarray, storage_options: Union[Dict[Any, Any], None] = None):
    """Write the object spatial distribution information to a healpix FITS file.

    Args:
        catalog_path (str): base path for the catalog
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        storage_options: dictionary that contains abstract filesystem credentials
    """
    catalog_base_dir = file_io.get_file_pointer_from_path(catalog_path)
    map_file_pointer = paths.get_point_map_file_pointer(catalog_base_dir)
    file_io.write_fits_image(histogram, map_file_pointer, storage_options=storage_options)



import json
import os

import pandas as pd
import pytest

from hipscat.catalog import CatalogType
from hipscat.catalog.association_catalog.association_catalog import AssociationCatalog
from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_node_type import PixelNodeType


def test_init_catalog(association_catalog_info, association_catalog_join_pixels):
    catalog = AssociationCatalog(
        association_catalog_info, [HealpixPixel(0, 11)], association_catalog_join_pixels
    )
    assert catalog.catalog_name == association_catalog_info.catalog_name
    pd.testing.assert_frame_equal(catalog.get_join_pixels(), association_catalog_join_pixels)
    assert catalog.get_healpix_pixels() == [HealpixPixel(0, 11)]
    assert catalog.catalog_info == association_catalog_info

    assert len(catalog.get_healpix_pixels()) == len([HealpixPixel(0, 11)])
    for hp_pixel in catalog.get_healpix_pixels():
        assert hp_pixel in [HealpixPixel(0, 11)]
        assert hp_pixel in catalog.pixel_tree
        assert catalog.pixel_tree[hp_pixel].node_type == PixelNodeType.LEAF


def test_wrong_catalog_type(association_catalog_info, association_catalog_join_pixels):
    association_catalog_info.catalog_type = CatalogType.OBJECT
    with pytest.raises(ValueError, match="catalog_type"):
        AssociationCatalog(association_catalog_info, [HealpixPixel(0, 11)], association_catalog_join_pixels)


def test_wrong_catalog_info_type(catalog_info, association_catalog_join_pixels):
    catalog_info.catalog_type = CatalogType.ASSOCIATION
    with pytest.raises(TypeError, match="catalog_info"):
        AssociationCatalog(catalog_info, [HealpixPixel(0, 11)], association_catalog_join_pixels)


def test_wrong_join_pixels_type(association_catalog_info):
    with pytest.raises(TypeError, match="join_pixels"):
        AssociationCatalog(association_catalog_info, [HealpixPixel(0, 11)], "test")


def test_different_join_pixels_type(association_catalog_info, association_catalog_join_pixels):
    partition_join_info = PartitionJoinInfo(association_catalog_join_pixels)
    catalog = AssociationCatalog(association_catalog_info, [HealpixPixel(0, 11)], partition_join_info)
    pd.testing.assert_frame_equal(catalog.get_join_pixels(), association_catalog_join_pixels)


def test_read_from_file(association_catalog_path, association_catalog_join_pixels):
    catalog = AssociationCatalog.read_from_hipscat(association_catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == association_catalog_path
    assert len(catalog.get_join_pixels()) == 4
    assert len(catalog.get_healpix_pixels()) == 1
    pd.testing.assert_frame_equal(catalog.get_join_pixels(), association_catalog_join_pixels)
    assert catalog.get_healpix_pixels() == [HealpixPixel(0, 11)]

    info = catalog.catalog_info
    assert info.primary_catalog == "small_sky"
    assert info.primary_column == "id"
    assert info.join_catalog == "small_sky_order1"
    assert info.join_column == "id"


def test_empty_directory(tmp_path, association_catalog_info_data, association_catalog_join_pixels):
    """Test loading empty or incomplete data"""
    ## Path doesn't exist
    with pytest.raises(FileNotFoundError):
        AssociationCatalog.read_from_hipscat(os.path.join("path", "empty"))

    catalog_path = os.path.join(tmp_path, "empty")
    os.makedirs(catalog_path, exist_ok=True)

    ## Path exists but there's nothing there
    with pytest.raises(FileNotFoundError):
        AssociationCatalog.read_from_hipscat(catalog_path)

    ## catalog_info file exists - getting closer
    file_name = os.path.join(catalog_path, "catalog_info.json")
    with open(file_name, "w", encoding="utf-8") as metadata_file:
        metadata_file.write(json.dumps(association_catalog_info_data))

    with pytest.raises(FileNotFoundError):
        AssociationCatalog.read_from_hipscat(catalog_path)

    ## partition_info file exists - almost there
    file_name = os.path.join(catalog_path, "partition_info.csv")
    with open(file_name, "w", encoding="utf-8") as metadata_file:
        metadata_file.write("Norder,Dir,Npix")

    with pytest.raises(FileNotFoundError):
        AssociationCatalog.read_from_hipscat(catalog_path)

    ## partition_join info file exists - enough to create a catalog
    file_name = os.path.join(catalog_path, "partition_join_info.csv")
    association_catalog_join_pixels.to_csv(file_name)

    catalog = AssociationCatalog.read_from_hipscat(catalog_path)
    assert catalog.catalog_name == association_catalog_info_data["catalog_name"]



"""Catalog Info for a HiPSCat Index table"""

from dataclasses import dataclass, field
from typing import List

from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo


@dataclass
class IndexCatalogInfo(BaseCatalogInfo):
    """Catalog Info for a HiPSCat Index table"""

    primary_catalog: str = None
    """Reference to object or source catalog"""

    indexing_column: str = None
    """Column that we provide an index over"""

    extra_columns: List[str] = field(default_factory=list)
    """Any additional payload columns included in index"""

    required_fields = BaseCatalogInfo.required_fields + [
        "primary_catalog",
        "indexing_column",
    ]

    DEFAULT_TYPE = CatalogType.INDEX
    REQUIRED_TYPE = CatalogType.INDEX



"""Test pixel path creation"""

import pytest

from hipscat.io import paths


def test_pixel_directory():
    """Simple case with sensical inputs"""
    expected = "/foo/Norder=0/Dir=0"
    result = paths.pixel_directory("/foo", 0, 5)
    assert result == expected


def test_pixel_directory_number():
    """Simple case with sensical inputs"""
    expected = "/foo/Norder=0/Dir=0"
    result = paths.pixel_directory("/foo", pixel_order=0, pixel_number=5, directory_number=0)
    assert result == expected

    result = paths.pixel_directory("/foo", pixel_order=0, directory_number=0)
    assert result == expected


def test_pixel_directory_missing():
    """Simple case with missing inputs"""
    with pytest.raises(ValueError):
        paths.pixel_directory("/foo", 0)


def test_pixel_directory_nonint():
    """Simple case with non-integer inputs"""
    with pytest.raises(ValueError):
        paths.pixel_directory("/foo", "zero", "five")


def test_pixel_catalog_file():
    """Simple case with sensical inputs"""
    expected = "/foo/Norder=0/Dir=0/Npix=5.parquet"
    result = paths.pixel_catalog_file("/foo", 0, 5)
    assert result == expected


def test_pixel_catalog_file_nonint():
    """Simple case with non-integer inputs"""
    with pytest.raises(ValueError):
        paths.pixel_catalog_file("/foo", "zero", "five")



import dataclasses
from typing import Any, Dict, Optional, Union

from hipscat.catalog.association_catalog.association_catalog_info import AssociationCatalogInfo
from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.catalog.index.index_catalog_info import IndexCatalogInfo
from hipscat.catalog.margin_cache.margin_cache_catalog_info import MarginCacheCatalogInfo
from hipscat.catalog.source_catalog.source_catalog_info import SourceCatalogInfo
from hipscat.io import FilePointer, file_io, paths

CATALOG_TYPE_TO_INFO_CLASS = {
    CatalogType.OBJECT: CatalogInfo,
    CatalogType.SOURCE: SourceCatalogInfo,
    CatalogType.ASSOCIATION: AssociationCatalogInfo,
    CatalogType.INDEX: IndexCatalogInfo,
    CatalogType.MARGIN: MarginCacheCatalogInfo,
}
"""Map of catalog types to their expected subclass of BaseCatalogInfo."""


def create_catalog_info(keywords: dict, catalog_type: Optional[CatalogType] = None) -> BaseCatalogInfo:
    """Generate a typed catalog info object from the type specified explicitly or
    using ``catalog_type`` keyword.

    Args:
        keywords: dictionary of catalog info keywords (e.g. from reading a
            ``catalog_info.json`` file).
        catalog_type: explicit request for a specific catalog type. if not
            provided, we will look for a key ``catalog_type`` in the keywords.
    Returns:
        populated BaseCatalogInfo of appropriate type.
    """

    if not catalog_type:
        if "catalog_type" not in keywords.keys():
            raise ValueError("catalog type is required to create catalog info object")
        catalog_type = keywords["catalog_type"]

    if catalog_type not in CatalogType.all_types():
        raise ValueError(f"Unknown catalog type: {catalog_type}")

    if catalog_type not in CATALOG_TYPE_TO_INFO_CLASS:  # pragma: no cover
        raise NotImplementedError(f"Unhandled catalog type: {catalog_type}")
    ci_class = CATALOG_TYPE_TO_INFO_CLASS[catalog_type]
    catalog_info_keywords = {}
    for field in dataclasses.fields(ci_class):
        if field.name in keywords:
            catalog_info_keywords[field.name] = keywords[field.name]
    return ci_class(**catalog_info_keywords)


def from_catalog_dir(catalog_base_dir: FilePointer, storage_options: Union[Dict[Any, Any], None] = None):
    """Generate a typed catalog info object from the type specified in the
    catalog info file.

    Args:
        catalog_base_dir: a path pointing to the base directory of a catalog,
            or may point to a ``catalog_info.json`` file directly.
        storage_options: dictionary that contains abstract filesystem credentials

    Returns:
        populated BaseCatalogInfo of appropriate type.
    """
    if file_io.is_regular_file(catalog_base_dir, storage_options=storage_options):
        ## This might be the catalog_info.json file - try anyway
        metadata_keywords = file_io.load_json_file(catalog_base_dir, storage_options=storage_options)
    else:
        catalog_info_file = paths.get_catalog_info_pointer(catalog_base_dir)
        metadata_keywords = file_io.load_json_file(catalog_info_file, storage_options=storage_options)
    return create_catalog_info(metadata_keywords)



from __future__ import annotations

from bisect import bisect
from typing import List

from hipscat.pixel_math import HealpixInputTypes, get_healpix_pixel
from hipscat.pixel_tree.pixel_node_type import PixelNodeType


class PixelNode:
    """A node in the HiPSCat quadtree of HEALPix pixels

    Attributes:
        node_type: If the node is a leaf, root, or inner type
        parent: The parent pixel node in the tree
        children: The child nodes of the pixel

    """

    _NODE_TYPE_MAX_CHILDREN = {
        PixelNodeType.ROOT: 12,
        PixelNodeType.INNER: 4,
        PixelNodeType.LEAF: 0,
    }

    def __init__(
        self,
        pixel: HealpixInputTypes,
        node_type: PixelNodeType,
        parent: PixelNode | None,
    ) -> None:
        """Inits PixelNode with its attributes

        Raises:
            ValueError: Invalid arguments for the specified pixel type
        """

        pixel = get_healpix_pixel(pixel)

        if node_type == PixelNodeType.ROOT:
            if parent is not None:
                raise ValueError("Root node cannot have a parent")
            if pixel.order != -1:
                raise ValueError("Root node must be at order -1")

        if node_type in (PixelNodeType.INNER, PixelNodeType.LEAF):
            if parent is None:
                raise ValueError("Inner and leaf nodes must have a parent")
            if pixel.order < 0 or pixel.pixel < 0:
                raise ValueError("Inner and leaf nodes must have an order and pixel number >= 0")

        if parent is not None and parent.hp_order != pixel.order - 1:
            raise ValueError("Parent node must be at order one less than current node")

        self.pixel = pixel
        self.node_type = node_type
        self.parent = parent
        self.children = []

        if self.parent is not None:
            self.parent.add_child_node(self)

    def __str__(self) -> str:
        return f"{self.node_type} Order: {self.pixel.order}, Pixel: {self.pixel.pixel}"

    def __repr__(self):
        return self.__str__()

    @property
    def hp_order(self):
        """The order of the HealpixPixel the node is at"""
        return self.pixel.order

    @property
    def hp_pixel(self):
        """The pixel number in NESTED ordering of the HealpixPixel the node is at"""
        return self.pixel.pixel

    def add_child_node(self, child: PixelNode):
        """Adds a child node to the node

        Maintains order of children based on pixel index

        Args:
            child: child node to add

        Raises:
            ValueError: The child to add does not have the current node set as a parent
            OverflowError: The node already has the maximum amount of children
        """
        if not child.parent == self:
            raise ValueError("Child node to add must have the node it is adding to as its parent")

        if len(self.children) >= self._NODE_TYPE_MAX_CHILDREN[self.node_type]:
            raise OverflowError("Node already has the maximum amount of children")

        insert_index = bisect(list(map(lambda node: node.hp_pixel, self.children)), child.hp_pixel)
        self.children.insert(insert_index, child)

    def remove_child_link(self, node):
        """Remove a link to a child node from the node. This removes it from the list of children,
        but does not delete the node

        Args:
            node: the node object to remove
        """
        if node not in self.children:
            raise ValueError("Node not a child node, cannot remove")
        self.children.remove(node)

    def get_all_leaf_descendants(self) -> List[PixelNode]:
        """Gets all descendant nodes that are leaf nodes.

        Leafs nodes correspond to pixels that have data files containing astronomical objects, so
        this method is used to get the files containing all astronomical objects within the pixel.

        This function called on a leaf node returns a list including the node itself and only itself

        Returns:
            A list containing all leaf nodes that descend from the current node
        """

        leaf_descendants = []
        self._add_all_leaf_descendants_rec(leaf_descendants)
        return leaf_descendants

    def _add_all_leaf_descendants_rec(self, leaf_descendants: List[PixelNode]):
        """Recursively add all leaf descendants to list

        list must be created outside function, done for efficiency vs list concat
        """
        if self.node_type == PixelNodeType.LEAF:
            leaf_descendants.append(self)

        for child in self.children:
            # pylint: disable=protected-access
            child._add_all_leaf_descendants_rec(leaf_descendants)



"""Utilities for reading and writing catalog files"""

from .file_io import FilePointer, get_file_pointer_from_path
from .paths import (
    create_hive_directory_name,
    create_hive_parquet_file_name,
    get_catalog_info_pointer,
    get_common_metadata_pointer,
    get_parquet_metadata_pointer,
    get_partition_info_pointer,
    get_point_map_file_pointer,
    get_provenance_pointer,
    pixel_catalog_file,
    pixel_directory,
)
from .write_metadata import (
    write_catalog_info,
    write_parquet_metadata,
    write_partition_info,
    write_provenance_info,
)



"""Catalog Info for a HiPSCat Margin cache table"""

from dataclasses import dataclass

from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo


@dataclass
class MarginCacheCatalogInfo(BaseCatalogInfo):
    """Catalog Info for a HiPSCat Margin Cache table"""

    primary_catalog: str = None
    """Reference to object or source catalog"""

    margin_threshold: float = None
    """Threshold of the pixel boundary, expressed in arcseconds."""

    required_fields = BaseCatalogInfo.required_fields + [
        "primary_catalog",
        "margin_threshold",
    ]

    DEFAULT_TYPE = CatalogType.MARGIN
    REQUIRED_TYPE = CatalogType.MARGIN



import os

import pytest

from hipscat.inspection.almanac import Almanac


def test_default(almanac_dir, test_data_dir):
    """Test loading from a default directory"""
    os.environ["HIPSCAT_DEFAULT_DIR"] = test_data_dir

    alms = Almanac(include_default_dir=True)
    assert len(alms.catalogs()) == 0

    os.environ["HIPSCAT_ALMANAC_DIR"] = almanac_dir
    alms = Almanac(include_default_dir=True)
    assert len(alms.catalogs()) == 8

    os.environ.pop("HIPSCAT_ALMANAC_DIR")
    alms = Almanac(include_default_dir=True)
    assert len(alms.catalogs()) == 0


def test_non_default(almanac_dir, test_data_dir):
    """Test loading with explicit (non-default) almanac base directory."""
    os.environ["HIPSCAT_DEFAULT_DIR"] = test_data_dir

    alms = Almanac(include_default_dir=False)
    assert len(alms.catalogs()) == 0

    alms = Almanac(include_default_dir=False, dirs=almanac_dir)
    assert len(alms.catalogs()) == 8

    alms = Almanac(include_default_dir=False, dirs=[almanac_dir])
    assert len(alms.catalogs()) == 8

    alms = Almanac(
        include_default_dir=False,
        dirs=[
            os.path.join(almanac_dir, "catalog.yml"),
            os.path.join(almanac_dir, "dataset.yml"),
        ],
    )
    assert len(alms.catalogs()) == 2


def test_namespaced(almanac_dir, test_data_dir):
    """Test that we can add duplicate catalogs, so long as we add a namespace."""
    os.environ["HIPSCAT_ALMANAC_DIR"] = almanac_dir
    os.environ["HIPSCAT_DEFAULT_DIR"] = test_data_dir

    with pytest.warns(match="Duplicate"):
        Almanac(include_default_dir=True, dirs=almanac_dir)

    alms = Almanac(
        include_default_dir=True,
        dirs={"custom": almanac_dir},
    )
    assert len(alms.catalogs()) == 16

    alms = Almanac(
        include_default_dir=False,
        dirs={"custom": almanac_dir, "custom2": almanac_dir},
    )
    assert len(alms.catalogs()) == 16


def test_catalogs_filters(default_almanac):
    """Test listing names of catalogs, using filters"""
    ## all (non-deprecated) catalogs
    assert len(default_almanac.catalogs()) == 8

    ## **all** catalogs
    assert len(default_almanac.catalogs(include_deprecated=True)) == 9

    ## all object and source (skip association/index/etc)
    assert len(default_almanac.catalogs(include_deprecated=True, types=["object", "source"])) == 6

    ## all active object and source
    assert len(default_almanac.catalogs(types=["object", "source"])) == 5

    ## non-existent type
    assert len(default_almanac.catalogs(types=["foo"])) == 0


def test_linked_catalogs_object(default_almanac):
    """Check that we can access the affiliated catalogs"""
    object_almanac = default_almanac.get_almanac_info("small_sky")
    assert len(object_almanac.sources) == 1

    source_almanac = object_almanac.sources[0]
    assert source_almanac.catalog_name == "small_sky_source_catalog"

    source_almanac = default_almanac.get_almanac_info(object_almanac.sources[0].catalog_name)
    assert source_almanac.catalog_name == "small_sky_source_catalog"

    source_catalog = default_almanac.get_catalog(object_almanac.sources[0].catalog_name)
    assert source_catalog.catalog_name == "small_sky_source_catalog"


def test_linked_catalogs_source(default_almanac, test_data_dir):
    """Check that we can access the affiliated catalogs"""
    source_almanac = default_almanac.get_almanac_info("small_sky_source_catalog")
    assert len(source_almanac.objects) == 1

    object_almanac = source_almanac.objects[0]
    assert object_almanac.catalog_name == "small_sky"

    source_almanac = default_almanac.get_almanac_info("small_sky_source_catalog")
    assert len(source_almanac.objects) == 1

    ## This source catalog has no object catalog, *and that's ok*
    new_almanac = Almanac(
        dirs=os.path.join(test_data_dir, "almanac_exception", "standalone_source_catalog.yml")
    )
    source_almanac = new_almanac.get_almanac_info("just_the_small_sky_source_catalog")
    assert len(source_almanac.objects) == 0


def test_linked_catalogs_association(default_almanac):
    """Check that read almanac entries are fully linked to one another."""

    association_almanac = default_almanac.get_almanac_info("small_sky_to_small_sky_order1")
    assert association_almanac.catalog_name == "small_sky_to_small_sky_order1"

    primary_almanac = association_almanac.primary_link
    assert primary_almanac
    assert primary_almanac.catalog_name == "small_sky"
    assert len(primary_almanac.associations) == 1
    assert len(primary_almanac.associations_right) == 0

    join_almanac = association_almanac.join_link
    assert join_almanac
    assert join_almanac.catalog_name == "small_sky_order1"
    assert len(join_almanac.associations) == 0
    assert len(join_almanac.associations_right) == 1


def test_linked_catalogs_index(default_almanac):
    """Check that read almanac entries are fully linked to one another."""

    index_almanac = default_almanac.get_almanac_info("index_catalog")
    assert index_almanac.catalog_name == "index_catalog"

    primary_almanac = index_almanac.primary_link
    assert primary_almanac
    assert primary_almanac.catalog_name == "catalog"
    assert len(primary_almanac.indexes) == 1


def test_linked_catalogs_margin(default_almanac):
    """Check that read almanac entries are fully linked to one another."""

    margin_almanac = default_almanac.get_almanac_info("margin_cache")
    assert margin_almanac.catalog_name == "margin_cache"

    primary_almanac = margin_almanac.primary_link
    assert primary_almanac
    assert primary_almanac.catalog_name == "catalog"
    assert len(primary_almanac.margins) == 1


def test_get_catalog(default_almanac):
    """Test that catalogs in almanac really exist (in test directory)"""

    for catalog_name in default_almanac.catalogs():
        catalog = default_almanac.get_catalog(catalog_name)
        assert catalog
        assert catalog.catalog_name == catalog_name


def test_get_catalog_exceptions(test_data_dir):
    """Test that we can create almanac entries, where catalogs might not exist."""
    bad_catalog_path_file = os.path.join(test_data_dir, "almanac_exception", "bad_catalog_path.yml")

    bad_links = Almanac(include_default_dir=False, dirs=bad_catalog_path_file)
    assert len(bad_links.catalogs()) == 1
    with pytest.raises(FileNotFoundError):
        bad_links.get_catalog("non_existent")


@pytest.mark.parametrize(
    "file_name,expected_error_match",
    [
        ("bad_type.yml", "foo"),
        (
            "association_missing_primary.yml",
            "association table .* missing primary catalog",
        ),
        (
            "association_missing_join.yml",
            "association table .* missing join catalog",
        ),
        (
            "index_missing_primary.yml",
            "index table .* missing primary catalog",
        ),
        (
            "margin_missing_primary.yml",
            "margin table .* missing primary catalog",
        ),
        (
            "bad_primary_path.yml",
            "source catalog .* missing object catalog /does/not/exist",
        ),
    ],
)
def test_almanac_creation(test_data_dir, file_name, expected_error_match):
    """Test that we throw exceptions, where bad almanac data or links exist in the files."""
    bad_links_file = os.path.join(test_data_dir, "almanac_exception", file_name)

    with pytest.warns(match=expected_error_match):
        Almanac(dirs=bad_links_file)



import pytest

from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_node import PixelNode


def check_node_data_correct(node, node_dict):
    assert node.pixel == node_dict["pixel"]
    assert node.node_type == node_dict["node_type"]
    assert node.parent == node_dict["parent"]


def test_create_root_node(root_pixel_node_data):
    root_node = PixelNode(**root_pixel_node_data)
    assert root_node
    check_node_data_correct(root_node, root_pixel_node_data)


def test_create_root_node_with_parent_raises_error(root_pixel_node, root_pixel_node_data):
    with pytest.raises(ValueError):
        root_pixel_node_data["parent"] = root_pixel_node
        PixelNode(**root_pixel_node_data)


def test_create_root_node_with_wrong_order_raises_error(root_pixel_node_data):
    with pytest.raises(ValueError):
        root_pixel_node_data["pixel"] = HealpixPixel(1, root_pixel_node_data["pixel"].pixel)
        PixelNode(**root_pixel_node_data)


def test_create_inner_node(inner_pixel_node_data, root_pixel_node):
    leaf_node = PixelNode(**inner_pixel_node_data)
    assert leaf_node
    check_node_data_correct(leaf_node, inner_pixel_node_data)
    assert len(root_pixel_node.children) == 1
    assert leaf_node in root_pixel_node.children


def test_create_inner_node_with_no_parent_raises_error(inner_pixel_node_data):
    with pytest.raises(ValueError):
        inner_pixel_node_data["parent"] = None
        PixelNode(**inner_pixel_node_data)


def test_create_leaf_node(leaf_pixel_node_data, inner_pixel_node):
    leaf_node = PixelNode(**leaf_pixel_node_data)
    assert leaf_node
    check_node_data_correct(leaf_node, leaf_pixel_node_data)
    assert len(inner_pixel_node.children) == 1
    assert leaf_node in inner_pixel_node.children


def test_create_leaf_node_with_no_parent_raises_error(leaf_pixel_node_data):
    with pytest.raises(ValueError):
        leaf_pixel_node_data["parent"] = None
        PixelNode(**leaf_pixel_node_data)


def test_create_leaf_node_with_negative_hp_order_raises_error(leaf_pixel_node_data):
    with pytest.raises(ValueError):
        leaf_pixel_node_data["pixel"] = HealpixPixel(-1, leaf_pixel_node_data["pixel"].pixel)
        PixelNode(**leaf_pixel_node_data)


def test_create_leaf_node_with_negative_hp_pixel_raises_error(leaf_pixel_node_data):
    with pytest.raises(ValueError):
        leaf_pixel_node_data["pixel"] = HealpixPixel(leaf_pixel_node_data["pixel"].order, -1)
        PixelNode(**leaf_pixel_node_data)


def test_create_node_with_wrong_order_parent_raises_error(leaf_pixel_node_data, root_pixel_node):
    with pytest.raises(ValueError):
        leaf_pixel_node_data["parent"] = root_pixel_node
        PixelNode(**leaf_pixel_node_data)


def test_add_child_node_with_wrong_child_parent_raises_error(leaf_pixel_node, root_pixel_node):
    with pytest.raises(ValueError):
        root_pixel_node.add_child_node(leaf_pixel_node)


def test_add_max_children_to_root_node(inner_pixel_node_data):
    inner_nodes = []
    for _ in range(12):
        inner_pixel_node_data["pixel"] = HealpixPixel(
            inner_pixel_node_data["pixel"].order,
            inner_pixel_node_data["pixel"].pixel + 1,
        )
        inner_nodes.append(PixelNode(**inner_pixel_node_data))


def test_add_too_many_children_to_root_node_fails(inner_pixel_node_data):
    with pytest.raises(OverflowError):
        inner_nodes = []
        for _ in range(13):
            inner_pixel_node_data["pixel"] = HealpixPixel(
                inner_pixel_node_data["pixel"].order,
                inner_pixel_node_data["pixel"].pixel + 1,
            )
            inner_nodes.append(PixelNode(**inner_pixel_node_data))


def test_add_max_children_to_inner_node(leaf_pixel_node_data):
    leaf_nodes = []
    for _ in range(4):
        leaf_pixel_node_data["pixel"] = HealpixPixel(
            leaf_pixel_node_data["pixel"].order, leaf_pixel_node_data["pixel"].pixel + 1
        )

        leaf_nodes.append(PixelNode(**leaf_pixel_node_data))


def test_add_too_many_children_to_inner_node_fails(leaf_pixel_node_data):
    with pytest.raises(OverflowError):
        leaf_nodes = []
        for _ in range(5):
            leaf_pixel_node_data["pixel"] = HealpixPixel(
                leaf_pixel_node_data["pixel"].order,
                leaf_pixel_node_data["pixel"].pixel + 1,
            )

            leaf_nodes.append(PixelNode(**leaf_pixel_node_data))


def test_add_child_to_leaf_node_fails(leaf_pixel_node, leaf_pixel_node_data):
    with pytest.raises(OverflowError):
        leaf_pixel_node_data["pixel"] = HealpixPixel(
            leaf_pixel_node.hp_order + 1, leaf_pixel_node_data["pixel"].pixel
        )

        leaf_pixel_node_data["parent"] = leaf_pixel_node
        PixelNode(**leaf_pixel_node_data)


def test_get_leaf_descendants_with_leaf_node(leaf_pixel_node):
    assert leaf_pixel_node.get_all_leaf_descendants() == [leaf_pixel_node]


def test_get_leaf_descendants_with_root_node(root_pixel_node, leaf_pixel_node):
    assert root_pixel_node.get_all_leaf_descendants() == [leaf_pixel_node]


def test_get_leaf_descendants_with_multiple_leafs(root_pixel_node, leaf_pixel_node, leaf_pixel_node_data):
    leaf_pixel_node_data["pixel"] = HealpixPixel(
        leaf_pixel_node_data["pixel"].order, leaf_pixel_node_data["pixel"].pixel + 1
    )
    second_leaf = PixelNode(**leaf_pixel_node_data)
    assert root_pixel_node.get_all_leaf_descendants() == [leaf_pixel_node, second_leaf]


def test_remove_child_link(root_pixel_node, inner_pixel_node):
    assert inner_pixel_node in root_pixel_node.children
    assert len(root_pixel_node.children) == 1
    root_pixel_node.remove_child_link(inner_pixel_node)
    assert inner_pixel_node not in root_pixel_node.children
    assert len(root_pixel_node.children) == 0


def test_remove_child_link_wrong_node_errors(root_pixel_node, leaf_pixel_node):
    assert leaf_pixel_node not in root_pixel_node.children
    with pytest.raises(ValueError):
        leaf_pixel_node.remove_child_link(leaf_pixel_node)


def test_node_string(root_pixel_node, leaf_pixel_node):
    root_string = "PixelNodeType.ROOT Order: -1, Pixel: -1"
    assert str(root_pixel_node) == root_string
    assert repr(root_pixel_node) == root_string

    leaf_string = "PixelNodeType.LEAF Order: 1, Pixel: 12"
    assert str(leaf_pixel_node) == leaf_string
    assert repr(leaf_pixel_node) == leaf_string



"""Utilities to build bounding boxes around healpixels that include a neighor margin."""

import healpy as hp
import numpy as np
from astropy.coordinates import SkyCoord
from numba import njit


def check_margin_bounds(r_asc, dec, pixel_order, pixel, margin_threshold, step=100, chunk_size=1000):
    # pylint: disable=too-many-locals
    """Check a set of points in ra and dec space to see if they fall within the
        `margin_threshold` of a given healpixel.

    Args:
        r_asc (:obj:`np.array`): one dimmensional array representing the ra of a
            given set of points.
        dec (:obj:`np.array`): one dimmensional array representing the dec of a
            given set of points.
        pixel_order (int): the order of the healpixel to be checked against.
        pixel (int): the healpixel to be checked against.
        margin_threshold (double): the size of the margin cache in arcseconds.
        step (int): the amount of samples we'll get from the healpixel boundary
            to test the datapoints against. the higher the step the more
            granular the point checking and therefore the more accurate, however
            using a smaller step value might be helpful for super large datasets
            to save compute time if accuracy is less important.
        chunk_size (int): the size of the chunk of data points we process on a single
            thread at any given time. We only process a certain amount of the data
            at once as trying to check all data points at once can lead to
            memory issues.
    Returns:
        :obj:`numpy.array` of boolean values corresponding to the ra and dec
            coordinates checked against whether a given point is within any of the
            provided polygons.
    """
    num_points = len(r_asc)
    if num_points != len(dec):
        raise ValueError(
            f"length of r_asc ({num_points}) is different \
                from length of dec ({len(dec)})"
        )

    if num_points == 0:
        return np.array([])

    bounds = hp.vec2dir(hp.boundaries(2**pixel_order, pixel, step=step, nest=True), lonlat=True)
    margin_check = np.full((num_points,), False)

    distances = _get_boundary_point_distances(pixel_order, pixel, step)
    margin_threshold_deg = margin_threshold / 3600.0

    num_bounds = step * 4
    ra_chunks = np.array_split(r_asc, chunk_size)
    dec_chunks = np.array_split(dec, chunk_size)
    index = 0
    for chunk_index, ra_chunk in enumerate(ra_chunks):
        dec_chunk = dec_chunks[chunk_index]

        chunk_len = len(ra_chunk)

        if chunk_len > 0:
            ra_matrix = np.repeat(np.reshape([ra_chunk], (-1, 1)), num_bounds, axis=1)
            dec_matrix = np.repeat(np.reshape([dec_chunk], (-1, 1)), num_bounds, axis=1)

            bounds_ra_matrix = np.repeat(np.array([bounds[0]]), len(ra_chunk), axis=0)
            bounds_dec_matrix = np.repeat(np.array([bounds[1]]), len(dec_chunk), axis=0)

            points_coords = SkyCoord(ra=ra_matrix, dec=dec_matrix, unit="deg")
            bounds_coords = SkyCoord(ra=bounds_ra_matrix, dec=bounds_dec_matrix, unit="deg")

            separations = points_coords.separation(bounds_coords)

            bisectors = np.apply_along_axis(
                _find_minimum_distance,
                1,
                separations,
                distances=distances,
                margin_threshold=margin_threshold_deg,
            )
            margin_check[index : index + chunk_len] = bisectors <= margin_threshold_deg
            index += chunk_len

            del ra_matrix, dec_matrix, bounds_ra_matrix, bounds_dec_matrix
            del points_coords, bounds_coords, separations

    del distances

    return margin_check


# numba jit compiler doesn't count for coverage tests, so we'll set no cover.
@njit("double(double[:], double[:], double)")
def _find_minimum_distance(separations, distances, margin_threshold):  # pragma: no cover
    """Find the minimum distance between a given datapoint and a healpixel"""
    minimum_index = np.argmin(separations)

    if separations[minimum_index] <= margin_threshold:
        return separations[minimum_index]

    plus_one, minus_one = minimum_index + 1, minimum_index - 1
    num_seps = len(separations)
    if minimum_index == 0:
        minus_one = num_seps - 1
    if minimum_index == num_seps - 1:
        plus_one = 0

    if separations[minus_one] <= separations[plus_one]:
        other_index = minus_one
    else:
        other_index = plus_one

    side_x = np.radians(separations[minimum_index])
    side_y = np.radians(separations[other_index])
    if (0 in (minimum_index, other_index)) and (num_seps - 1 in (minimum_index, other_index)):
        side_z = distances[-1]
    else:
        side_z = distances[min(minimum_index, other_index)]
    side_z *= np.pi / 180.0

    ang = np.arccos((np.cos(side_y) - (np.cos(side_x) * np.cos(side_z))) / (np.sin(side_x) * np.sin(side_z)))
    bisector = np.degrees(np.arcsin(np.sin(ang) * np.sin(side_x)))

    return bisector


def _get_boundary_point_distances(order, pixel, step):
    """Get the distance of the segments between healpixel boundary points"""
    boundary_points = hp.vec2dir(hp.boundaries(2**order, pixel, step=step, nest=True), lonlat=True)

    # shift forward all the coordinates by one
    shift_ra = np.roll(boundary_points[0], 1)
    shift_dec = np.roll(boundary_points[1], 1)

    coord_set_1 = SkyCoord(ra=boundary_points[0], dec=boundary_points[1], unit="deg")
    coord_set_2 = SkyCoord(ra=shift_ra, dec=shift_dec, unit="deg")

    separations = coord_set_1.separation(coord_set_2).degree
    return separations



import dataclasses
import os

import pytest

from hipscat.inspection.almanac import Almanac
from hipscat.inspection.almanac_info import AlmanacInfo


def test_from_catalog_dir(small_sky_dir):
    """Load from a directory."""
    almanac_info = AlmanacInfo.from_catalog_dir(small_sky_dir)
    assert almanac_info.catalog_name == "small_sky"


def test_from_bad_file(association_catalog_partition_join_file):
    """Test failures when loading almanac from invalid files."""
    with pytest.raises(ValueError, match="csv"):
        AlmanacInfo.from_file(association_catalog_partition_join_file)


def test_write_to_file(tmp_path, association_catalog_path):
    """Write out the almanac to file and make sure we can read it again."""
    almanac_info = AlmanacInfo.from_catalog_dir(association_catalog_path)
    assert almanac_info.catalog_name == "small_sky_to_small_sky_order1"
    almanac_info.version = "v0.0.1"
    almanac_info.deprecated = "yes - use something else"
    almanac_info.creators.append("me")

    almanac_info.write_to_file(tmp_path, default_dir=False)

    new_info = AlmanacInfo.from_file(os.path.join(tmp_path, "small_sky_to_small_sky_order1.yml"))

    assert new_info.catalog_name == almanac_info.catalog_name

    almanac_info_dict = dataclasses.asdict(almanac_info)
    new_info_dict = dataclasses.asdict(new_info)
    for key, value in new_info_dict.items():
        assert almanac_info_dict[key] == value


def test_write_to_file_load_almanac(tmp_path, small_sky_dir):
    """Write out the almanac to file and make sure we can read it again."""
    almanac_info = AlmanacInfo.from_catalog_dir(small_sky_dir)
    assert almanac_info.catalog_name == "small_sky"

    almanac_info.write_to_file(tmp_path, default_dir=False)

    alms = Almanac(include_default_dir=False, dirs=tmp_path)
    assert len(alms.catalogs()) == 1


def test_write_to_bad_file(tmp_path, small_sky_dir):
    """Test failure conditions when writing almanac."""
    os.environ["HIPSCAT_ALMANAC_DIR"] = str(tmp_path)
    almanac_info = AlmanacInfo.from_catalog_dir(small_sky_dir)

    with pytest.raises(ValueError, match="only one"):
        almanac_info.write_to_file(tmp_path, default_dir=True)

    with pytest.raises(ValueError, match="foo"):
        almanac_info.write_to_file(fmt="foo")

    almanac_info.write_to_file(default_dir=True)
    with pytest.raises(ValueError, match="already exists"):
        almanac_info.write_to_file(default_dir=True)


def test_association_fields(association_catalog_path, index_catalog_info_file, small_sky_dir):
    """Test additional text fields tables with primary/join relationships."""
    almanac_info = AlmanacInfo.from_catalog_dir(association_catalog_path)
    assert almanac_info.catalog_name == "small_sky_to_small_sky_order1"
    assert almanac_info.primary == "small_sky"
    assert almanac_info.join == "small_sky_order1"

    almanac_info = AlmanacInfo.from_catalog_dir(index_catalog_info_file)
    assert almanac_info.primary == "catalog"
    assert almanac_info.join is None

    almanac_info = AlmanacInfo.from_catalog_dir(small_sky_dir)
    assert almanac_info.primary is None
    assert almanac_info.join is None


## Commented out -
## As new test catalogs are added, you can use this method to quickly add
## their almanac to the default almanac directory
# def test_add_almanac(source_catalog_info_file):
#     """Write out the almanac to file and make sure we can read it again."""
#     almanac_info = AlmanacInfo.from_catalog_dir(source_catalog_info_file)
#     almanac_info.write_to_file()



"""Tests of partition info functionality"""
import os

import pytest

from hipscat.catalog import PartitionInfo
from hipscat.io import file_io, paths
from hipscat.pixel_math import HealpixPixel


def test_load_partition_info_small_sky(small_sky_dir):
    """Instantiate the partition info for catalog with 1 pixel"""
    partition_info_file = paths.get_partition_info_pointer(small_sky_dir)
    partitions = PartitionInfo.read_from_file(partition_info_file)

    order_pixel_pairs = partitions.get_healpix_pixels()
    assert len(order_pixel_pairs) == 1
    expected = [HealpixPixel(0, 11)]
    assert order_pixel_pairs == expected


def test_load_partition_info_small_sky_order1(small_sky_order1_dir):
    """Instantiate the partition info for catalog with 4 pixels"""
    partition_info_file = paths.get_partition_info_pointer(small_sky_order1_dir)
    partitions = PartitionInfo.read_from_file(partition_info_file)

    order_pixel_pairs = partitions.get_healpix_pixels()
    assert len(order_pixel_pairs) == 4
    expected = [
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
        HealpixPixel(1, 47),
    ]
    assert order_pixel_pairs == expected


def test_load_partition_no_file(tmp_path):
    wrong_path = os.path.join(tmp_path, "wrong.csv")
    wrong_pointer = file_io.get_file_pointer_from_path(wrong_path)
    with pytest.raises(FileNotFoundError):
        PartitionInfo.read_from_file(wrong_pointer)


def test_get_highest_order(small_sky_order1_dir):
    """test the `get_highest_order` method"""
    partition_info_file = paths.get_partition_info_pointer(small_sky_order1_dir)
    partitions = PartitionInfo.read_from_file(partition_info_file)

    highest_order = partitions.get_highest_order()

    assert highest_order == 1


def test_write_to_file(tmp_path, small_sky_pixels):
    """Write out the partition info to file and make sure we can read it again."""
    partition_info_pointer = paths.get_partition_info_pointer(tmp_path)
    partition_info = PartitionInfo.from_healpix(small_sky_pixels)
    partition_info.write_to_file(partition_info_pointer)

    new_partition_info = PartitionInfo.read_from_file(partition_info_pointer)

    assert partition_info.get_healpix_pixels() == new_partition_info.get_healpix_pixels()



from hipscat.catalog.dataset.catalog_info_factory import from_catalog_dir
from hipscat.io import get_partition_info_pointer
from hipscat.io.file_io.file_pointer import FilePointer, is_regular_file


def is_valid_catalog(pointer: FilePointer) -> bool:
    """Checks if a catalog is valid for a given base catalog pointer

    Args:
        pointer: pointer to base catalog directory

    Returns:
        True if both the catalog_info and partition_info files are
        valid, False otherwise
    """
    return is_catalog_info_valid(pointer) and is_partition_info_valid(pointer)


def is_catalog_info_valid(pointer):
    """Checks if catalog_info is valid for a given base catalog pointer

    Args:
        pointer: pointer to base catalog directory

    Returns:
        True if the catalog_info file exists, and it is correctly formatted,
        False otherwise
    """
    is_valid = True
    try:
        from_catalog_dir(pointer)
    except (FileNotFoundError, ValueError, NotImplementedError):
        is_valid = False
    return is_valid


def is_partition_info_valid(pointer):
    """Checks if partition_info is valid for a given base catalog pointer

    Args:
        pointer: pointer to base catalog directory

    Returns:
        True if the partition_info file exists, False otherwise
    """
    partition_info_pointer = get_partition_info_pointer(pointer)
    partition_info_exists = is_regular_file(partition_info_pointer)
    return partition_info_exists



import os

from hipscat.io.file_io.file_io import get_fs


def copy_tree_fs_to_fs(
    fs1_source: str,
    fs2_destination: str,
    storage_options1: dict = None,
    storage_options2: dict = None,
    verbose=False,
):
    """Recursive Copies directory from one filesystem to the other.

    Args:
        fs1_source: location of source directory to copy
        fs2_destination: location of destination directory to for fs1 to be written two
        storage_options1: dictionary that contains abstract filesystem1 credentials
        storage_options2: dictionary that contains abstract filesystem2 credentials
    """

    source_fs, source_fp = get_fs(fs1_source, storage_options=storage_options1)
    destination_fs, desintation_fp = get_fs(fs2_destination, storage_options=storage_options2)
    copy_dir(source_fs, source_fp, destination_fs, desintation_fp, verbose=verbose)


def copy_dir(source_fs, source_fp, destination_fs, desintation_fp, verbose=False, chunksize=1024 * 1024):
    """Recursive method to copy directories and their contents.

    Args:
        fs1: fsspec.filesystem for the source directory contents
        fs1_pointer: source directory to copy content files
        fs2: fsspec.filesytem for destination directory
        fs2_pointer: destination directory for copied contents
    """
    destination_folder = os.path.join(desintation_fp, source_fp.split("/")[-1])
    if destination_folder[-1] != "/":
        destination_folder += "/"
    if not destination_fs.exists(destination_folder):
        if verbose:
            print(f"Creating destination folder: {destination_folder}")
        destination_fs.makedirs(destination_folder, exist_ok=True)

    dir_contents = source_fs.listdir(source_fp)
    files = [x for x in source_fs.listdir(source_fp) if x["type"] == "file"]

    for _file in files:
        destination_fname = os.path.join(destination_folder, _file["name"].split("/")[-1])
        if verbose:
            print(f'Copying file {_file["name"]} to {destination_fname}')
        with source_fs.open(_file["name"], "rb") as source_file:
            with destination_fs.open(destination_fname, "wb") as destination_file:
                while True:
                    chunk = source_file.read(chunksize)
                    if not chunk:
                        break
                    destination_file.write(chunk)

    dirs = [x for x in dir_contents if x["type"] == "directory"]
    for _dir in dirs:
        copy_dir(
            source_fs, _dir["name"], destination_fs, destination_folder, chunksize=chunksize, verbose=verbose
        )


if __name__ == "__main__":

    source_pw = f"{os.getcwd()}/../tests/data"
    target_pw = "new_protocol:///path/to/pytest/hipscat"

    target_so = {
        "valid_storage_option_param1": os.environ.get("NEW_PROTOCOL_PARAM1"),
        "valid_storage_option_param1": os.environ.get("NEW_PROTOCOL_PARAM2"),
    }
    copy_tree_fs_to_fs(source_pw, target_pw, {}, target_so, verbose=True)



"""Set of convenience methods for testing file contents"""

import re

import pyarrow as pa
import pytest

from hipscat.io.file_io.file_io import load_text_file
from hipscat.io.file_io.file_pointer import does_file_or_directory_exist

# pylint: disable=missing-function-docstring, redefined-outer-name


@pytest.fixture
def assert_text_file_matches():
    def assert_text_file_matches(expected_lines, file_name, storage_options: dict = None):
        """Convenience method to read a text file and compare the contents, line for line.

        When file contents get even a little bit big, it can be difficult to see
        the difference between an actual file and the expected contents without
        increased testing verbosity. This helper compares files line-by-line,
        using the provided strings or regular expressions.

        Notes:
            Because we check strings as regular expressions, you may need to escape some
            contents of `expected_lines`.

        Args:
            expected_lines(:obj:`string array`) list of strings, formatted as regular expressions.
            file_name (str): fully-specified path of the file to read
            storage_options (dict): dictionary of filesystem storage options
        """
        assert does_file_or_directory_exist(
            file_name, storage_options=storage_options
        ), f"file not found [{file_name}]"
        contents = load_text_file(file_name, storage_options=storage_options)

        assert len(expected_lines) == len(
            contents
        ), f"files not the same length ({len(contents)} vs {len(expected_lines)})"
        for i, expected in enumerate(expected_lines):
            assert re.match(expected, contents[i]), (
                f"files do not match at line {i+1} " f"(actual: [{contents[i]}] vs expected: [{expected}])"
            )

    return assert_text_file_matches


@pytest.fixture
def basic_catalog_parquet_metadata():
    return pa.schema(
        [
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.int64()),
            pa.field("dec_error", pa.int64()),
            pa.field("__index_level_0__", pa.int64()),
        ]
    )



from dataclasses import dataclass

from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo


@dataclass
class CatalogInfo(BaseCatalogInfo):
    """Catalog Info for a HEALPix Hive partitioned Catalog"""

    epoch: str = "J2000"
    ra_column: str = "ra"
    dec_column: str = "dec"

    required_fields = BaseCatalogInfo.required_fields + [
        "epoch",
        "ra_column",
        "dec_column",
    ]

    DEFAULT_TYPE = CatalogType.OBJECT



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import List

from hipscat.pixel_math.hipscat_id import HIPSCAT_ID_HEALPIX_ORDER


@dataclass(eq=True, frozen=True, order=True)
class HealpixPixel:
    """A HEALPix pixel, represented by an order and pixel number in NESTED ordering scheme

    see https://lambda.gsfc.nasa.gov/toolbox/pixelcoords.html for more information
    """

    order: int
    pixel: int

    def __post_init__(self) -> None:
        """Initialize a HEALPix pixel
        Args:
            order: HEALPix order
            pixel: HEALPix pixel number in NESTED ordering scheme
        """
        if self.order > HIPSCAT_ID_HEALPIX_ORDER:
            raise ValueError(f"HEALPix order cannot be greater than {HIPSCAT_ID_HEALPIX_ORDER}")

    def __str__(self) -> str:
        return f"Order: {self.order}, Pixel: {self.pixel}"

    def __repr__(self):
        return self.__str__()

    def convert_to_lower_order(self, delta_order: int) -> HealpixPixel:
        """Returns the HEALPix pixel that contains the pixel at a lower order

        Args:
            delta_order: the difference in order to be subtracted from the current order to generate
                the pixel at a lower order. Must be non-negative

        Returns:
            A new `HealpixPixel` at the current order - `delta_order` which contains the current
            pixel

        Raises:
            ValueError: If delta_order is greater than the current order, a pixel cannot be
                generated at a negative order. Or if delta_order is negative

        """
        if self.order - delta_order < 0:
            raise ValueError("Pixel Order cannot be below zero")
        if delta_order < 0:
            raise ValueError("delta order cannot be below zero")
        new_order = self.order - delta_order
        new_pixel = math.floor(self.pixel / 4**delta_order)
        return HealpixPixel(new_order, new_pixel)

    def convert_to_higher_order(self, delta_order: int) -> List[HealpixPixel]:
        """Returns a list of HEALPix pixels making up the current pixel at a higher order

        Args:
            delta_order: the difference in order to be added to the current order to generate the
                pixels at a higher order. Must be non-negative

        Returns:
            A new `HealpixPixel` at the current order - `delta_order` which contains the current
            pixel

        Raises:
            ValueError: If delta_order + current order is greater than the maximum HEALPix order,
                pixels cannot be  generated. Or if delta_order is negative
        """
        if self.order + delta_order > HIPSCAT_ID_HEALPIX_ORDER:
            raise ValueError(f"Pixel Order cannot be above maximum order {HIPSCAT_ID_HEALPIX_ORDER}")
        if delta_order < 0:
            raise ValueError("delta order cannot be below zero")
        pixels = []
        new_order = self.order + delta_order
        for new_pixel in range(self.pixel * 4**delta_order, (self.pixel + 1) * 4**delta_order):
            pixels.append(HealpixPixel(new_order, new_pixel))
        return pixels



"""
Compute the hipscat index field

This index is defined as a 64-bit integer which has two parts:
    - healpix pixel (at order 19)
    - incrementing counter (within same healpix, for uniqueness)
    
::

    |------------------------------------------|-------------------|
    |<-----    healpixel at order 19    ------>|<--   counter   -->|

This provides us with an increasing index, that will not overlap
between spatially partitioned data files.

See the README in this directory for more information on the fiddly pixel math.
"""

import healpy as hp
import numpy as np

HIPSCAT_ID_COLUMN = "_hipscat_index"
HIPSCAT_ID_HEALPIX_ORDER = 19


def compute_hipscat_id(ra_values, dec_values):
    """Compute the hipscat ID field.

    Args:
        ra_values (list[float]): celestial coordinates, right ascension
        dec_values (list[float]): celestial coordinates, declination
    Returns:
        one-dimensional array of int64s with hipscat IDs for the sky positions.
    Raises:
        ValueError: if the length of the input lists don't match.
    """
    if len(ra_values) != len(dec_values):
        raise ValueError("ra and dec arrays should have the same length")

    ## Construct the bit-shifted healpix segment
    value_count = len(ra_values)
    mapped_pixels = hp.ang2pix(2**HIPSCAT_ID_HEALPIX_ORDER, ra_values, dec_values, nest=True, lonlat=True)

    ## We sort to put pixels next to each other that will need to be counted.
    ## This simplifies the counter logic, as we can subtract the index where
    ## we first see the pixel value from the current index to get the offset counter.
    sort_index = np.argsort(mapped_pixels, kind="stable")
    mapped_pixels = mapped_pixels[sort_index]

    ## Construct the counter.
    _, unique_indices, unique_inverse = np.unique(mapped_pixels, return_inverse=True, return_index=True)
    unique_indices = unique_indices.astype(np.uint64)
    boring_number_index = np.arange(value_count, dtype=np.uint64)
    offset_counter = boring_number_index - unique_indices[unique_inverse]

    ## Add counter to shifted pixel, and map back to the original, unsorted, values
    shifted_pixels = _compute_hipscat_id_from_mapped_pixels(mapped_pixels, offset_counter)

    unsort_index = np.argsort(sort_index, kind="stable")
    return shifted_pixels[unsort_index]


def _compute_hipscat_id_from_mapped_pixels(mapped_pixels, offset_counter):
    shifted_pixels = mapped_pixels.astype(np.uint64) << np.uint64(64 - (4 + 2 * HIPSCAT_ID_HEALPIX_ORDER))
    shifted_pixels = shifted_pixels + offset_counter
    return shifted_pixels


def hipscat_id_to_healpix(ids, target_order = HIPSCAT_ID_HEALPIX_ORDER):
    """Convert some hipscat ids to the healpix pixel at the specified order
    This is just bit-shifting the counter away.

    Args:
        ids (list[int64]): list of well-formatted hipscat ids
    Returns:
        list of target_order pixels from the hipscat id
    """
    return np.asarray(ids, dtype=np.uint64) >> (64 - (4 + 2 * target_order))


def healpix_to_hipscat_id(order: int, pixel: int, counter: int = 0) -> int:
    """Convert a healpix pixel to a hipscat_id

    This maps the healpix pixel to the lowest pixel number within that pixel at order 19,
    then shifts and adds the given counter to get a hipscat_id.

    Useful for operations such as filtering by hipscat_id.

    Args:
        order (int64 | list[int64]): order of pixel to convert
        pixel (int64 | list[int64]): pixel number in nested ordering of pixel to convert
        counter (int64 | list[int64]) (Default: 0): counter value in converted hipscat id
    Returns:
        list of hipscat ids
    """
    order = np.uint64(order)
    pixel = np.uint64(pixel)
    counter = np.uint64(counter)
    pixel_higher_order = pixel * (4 ** (HIPSCAT_ID_HEALPIX_ORDER - order))
    hipscat_id = _compute_hipscat_id_from_mapped_pixels(pixel_higher_order, counter)
    return hipscat_id



"""Catalog Info for a HiPSCat Source (detection/timeseries) table"""

from dataclasses import dataclass

from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.catalog.catalog_type import CatalogType


@dataclass
class SourceCatalogInfo(CatalogInfo):
    """Catalog Info for a HiPSCat Source (detection/timeseries) table.

    Includes some optional specification for timeseries-level columns.
    """

    primary_catalog: str = None
    """Object catalog reference"""

    mjd_column: str = ""
    """Column name for time of observation"""

    band_column: str = ""
    """Column name for photometric band"""

    mag_column: str = ""
    """Column name for magnitude measurement"""

    mag_err_column: str = ""
    """Column name for error in magnitude measurement"""

    DEFAULT_TYPE = CatalogType.SOURCE
    REQUIRED_TYPE = CatalogType.SOURCE

    ## NB: No additional required columns.



from __future__ import annotations

from typing import List

from hipscat.pixel_math import HealpixInputTypes, get_healpix_pixel
from hipscat.pixel_tree.pixel_node import PixelNode
from hipscat.pixel_tree.pixel_node_type import PixelNodeType


class PixelTree:
    """Sparse Quadtree of HEALPix pixels that make up the HiPSCat catalog

    This class stores each node in the tree, with leaf nodes corresponding to pixels with data
    files.

    There are a number of methods in this class which allow for quickly navigating through the
    tree and performing operations to filter the pixels in the catalog.

    Attributes:
        pixels: Nested dictionary of pixel nodes stored in the tree. Indexed by HEALPix
            order then pixel number
        root_pixel: Root node of the tree. Its children are a subset of the
            12 base HEALPix pixels
    """

    def __init__(self, root_pixel: PixelNode, pixels: dict[int, dict[int, PixelNode]]) -> None:
        """Initialises a tree object from the nodes in the tree

        Args:
            root_pixel: PixelNode representing the root node in the tree
            pixels: Dictionary containing all PixelNodes in the tree
        """
        self.root_pixel = root_pixel
        self.pixels = pixels

    def __len__(self):
        """Gets the number of nodes in the tree, including inner nodes and the root node

        Returns:
            The number of nodes in the tree, including inner nodes and the root node
        """
        pixel_count = 0
        for _, order_pixels in self.pixels.items():
            pixel_count += len(order_pixels)
        return pixel_count

    def contains(self, pixel: HealpixInputTypes) -> bool:
        """Check if tree contains a node at a given order and pixel

        Args:
            pixel: HEALPix pixel to check. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            True if the tree contains the pixel, False if not
        """
        pixel = get_healpix_pixel(pixel)
        return pixel.order in self.pixels and pixel.pixel in self.pixels[pixel.order]

    def __contains__(self, item):
        return self.contains(item)

    def get_node(self, pixel: HealpixInputTypes) -> PixelNode | None:
        """Get the node at a given pixel

        Args:
            pixel: HEALPix pixel to get. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            The PixelNode at the index, or None if a node does not exist
        """
        pixel = get_healpix_pixel(pixel)
        if self.contains(pixel):
            return self.pixels[pixel.order][pixel.pixel]
        return None

    def __getitem__(self, item):
        return self.get_node(item)

    def get_leaf_nodes_at_healpix_pixel(self, pixel: HealpixInputTypes) -> List[PixelNode]:
        """Lookup all leaf nodes that contain or are within a given HEALPix pixel

        - Exact matches will return a list with only the matching pixel
        - A pixel that is within a lower order pixel in the tree will return a list with the lower
          order pixel
        - A pixel that has higher order pixels within found in the tree will return a list with all
          higher order pixels
        - A pixel with no matching leaf nodes in the tree will return an empty list

        Args:
            pixel: HEALPix pixel of the pixel to lookup

        Returns:
            A list of the leaf PixelNodes in the tree that align with the specified pixel
        """
        pixel = get_healpix_pixel(pixel)

        if self.contains(pixel):
            # Pixel exists in tree. Either a leaf node with an exact match for the search pixel,
            # or an inner node, so the search pixel will contain leaf nodes at higher orders
            node_in_tree = self.get_node(pixel)
            return node_in_tree.get_all_leaf_descendants()
        # if the pixel doesn't exist in the tree, it's either because the tree doesn't cover the
        # pixel, or the pixel is at a higher order than the tree at that location, so we search for
        # lower order nodes in the tree
        node_in_tree = self._find_first_lower_order_leaf_node_in_tree(pixel)
        if node_in_tree is None:
            return []
        return [node_in_tree]

    def _find_first_lower_order_leaf_node_in_tree(self, pixel: HealpixInputTypes) -> PixelNode | None:
        pixel = get_healpix_pixel(pixel)
        for delta_order in range(1, pixel.order + 1):
            lower_pixel = pixel.convert_to_lower_order(delta_order)
            if self.contains(lower_pixel):
                lower_node = self.get_node(lower_pixel)
                if lower_node.node_type == PixelNodeType.LEAF:
                    # If the catalog doesn't fully cover the sky, it's possible we encounter an
                    # inner node whose leaf children don't cover the search pixel.
                    return lower_node
                return None
        return None



"""Utilities for generating and manipulating object count histograms"""

import healpy as hp
import numpy as np
import pandas as pd

from hipscat.pixel_math.healpix_pixel import HealpixPixel


def empty_histogram(highest_order):
    """Use numpy to create an histogram array with the right shape, filled with zeros.

    Args:
        highest_order (int): the highest healpix order (e.g. 0-10)
    Returns:
        one-dimensional numpy array of long integers, where the length is equal to
        the number of pixels in a healpix map of target order, and all values are set to 0.
    """
    return np.zeros(hp.order2npix(highest_order), dtype=np.int64)


def generate_histogram(
    data: pd.DataFrame,
    highest_order,
    ra_column="ra",
    dec_column="dec",
):
    """Generate a histogram of counts for objects found in `data`

    Args:
        data (:obj:`pd.DataFrame`): tabular object data
        highest_order (int):  the highest healpix order (e.g. 0-10)
        ra_column (str): where in the input to find the celestial coordinate, right ascension
        dec_column (str): where in the input to find the celestial coordinate, declination
    Returns:
        one-dimensional numpy array of long integers where the value at each index corresponds
        to the number of objects found at the healpix pixel.
    Raises:
        ValueError: if the `ra_column` or `dec_column` cannot be found in the input file.
    """
    histogram_result = empty_histogram(highest_order)

    # Verify that the data frame has columns with desired names.
    required_columns = [ra_column, dec_column]
    if not all(x in data.columns for x in required_columns):
        raise ValueError(f"Invalid column names in input: {ra_column}, {dec_column}")
    mapped_pixels = hp.ang2pix(
        2**highest_order,
        data[ra_column].values,
        data[dec_column].values,
        lonlat=True,
        nest=True,
    )
    mapped_pixel, count_at_pixel = np.unique(mapped_pixels, return_counts=True)
    histogram_result[mapped_pixel] += count_at_pixel.astype(np.int64)
    return histogram_result


def generate_alignment(histogram, highest_order=10, lowest_order=0, threshold=1_000_000):
    """Generate alignment from high order pixels to those of equal or lower order

    We may initially find healpix pixels at order 10, but after aggregating up to the pixel
    threshold, some final pixels are order 4 or 7. This method provides a map from pixels
    at order 10 to their destination pixel. This may be used as an input into later partitioning
    map reduce steps.

    Args:
        histogram (:obj:`np.array`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        highest_order (int):  the highest healpix order (e.g. 5-10)
        lowest_order (int): the lowest healpix order (e.g. 1-5). specifying a lowest order
            constrains the partitioning to prevent spatially large pixels.
        threshold (int): the maximum number of objects allowed in a single pixel
    Returns:
        one-dimensional numpy array of integer 3-tuples, where the value at each index corresponds
        to the destination pixel at order less than or equal to the `highest_order`.

        The tuple contains three integers:

        - order of the destination pixel
        - pixel number *at the above order*
        - the number of objects in the pixel
    Raises:
        ValueError: if the histogram is the wrong size, or some initial histogram bins
            exceed threshold.
    """
    if len(histogram) != hp.order2npix(highest_order):
        raise ValueError("histogram is not the right size")
    if lowest_order > highest_order:
        raise ValueError("lowest_order should be less than highest_order")
    max_bin = np.amax(histogram)
    if max_bin > threshold:
        raise ValueError(f"single pixel count {max_bin} exceeds threshold {threshold}")

    nested_sums = []
    for i in range(0, highest_order):
        nested_sums.append(empty_histogram(i))
    nested_sums.append(histogram)

    # work backward - from highest order, fill in the sums of lower order pixels
    for read_order in range(highest_order, lowest_order, -1):
        parent_order = read_order - 1
        for index in range(0, len(nested_sums[read_order])):
            parent_pixel = index >> 2
            nested_sums[parent_order][parent_pixel] += nested_sums[read_order][index]

    nested_alignment = []
    for i in range(0, highest_order + 1):
        nested_alignment.append(np.full(hp.order2npix(i), None))

    # work forward - determine if we should map to a lower order pixel, this pixel, or keep looking.
    for read_order in range(lowest_order, highest_order + 1):
        parent_order = read_order - 1
        for index in range(0, len(nested_sums[read_order])):
            parent_alignment = None
            if parent_order >= 0:
                parent_pixel = index >> 2
                parent_alignment = nested_alignment[parent_order][parent_pixel]

            if parent_alignment:
                nested_alignment[read_order][index] = parent_alignment
            elif nested_sums[read_order][index] == 0:
                continue
            elif nested_sums[read_order][index] <= threshold:
                nested_alignment[read_order][index] = (
                    read_order,
                    index,
                    nested_sums[read_order][index],
                )

    return nested_alignment[highest_order]


def generate_destination_pixel_map(histogram, pixel_map):
    """Generate mapping from destination pixel to all the constituent pixels.

    Args:
        histogram (:obj:`np.array`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        pixel_map (:obj:`np.array`): one-dimensional numpy array of integer 3-tuples.
            See :func:`generate_alignment` for more details on this format.
    Returns:
        dictionary that maps the integer 3-tuple of a pixel at destination order to the set of
        indexes in histogram for the pixels at the original healpix order
    """

    # Find all distinct destination pixels
    non_none_elements = pixel_map[pixel_map != np.array(None)]
    unique_pixels = np.unique(non_none_elements)

    # Compute the order from the number of pixels at the level.
    max_order = hp.npix2order(len(histogram))

    result = {}
    for order, pixel, count in unique_pixels:
        # Find all constituent pixels at the histogram's order
        explosion_factor = 4 ** (max_order - int(order))
        start_pixel = int(pixel) * explosion_factor
        end_pixel = (int(pixel) + 1) * explosion_factor

        non_zero_indexes = np.nonzero(histogram[start_pixel:end_pixel])[0] + start_pixel
        result[(order, pixel, count)] = non_zero_indexes.tolist()

    return result


def compute_pixel_map(histogram, highest_order=10, lowest_order=0, threshold=1_000_000):
    # pylint: disable=too-many-locals
    """Generate alignment from high order pixels to those of equal or lower order

    We may initially find healpix pixels at order 10, but after aggregating up to the pixel
    threshold, some final pixels are order 4 or 7. This method provides a map from destination
    healpix pixels at a lower order to the origin pixels at the highest order. This may be used
    as an input into later partitioning map reduce steps.

    Args:
        histogram (:obj:`np.array`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        highest_order (int):  the highest healpix order (e.g. 0-10)
        lowest_order (int): the lowest healpix order (e.g. 1-5). specifying a lowest order
            constrains the partitioning to prevent spatially large pixels.
        threshold (int): the maximum number of objects allowed in a single pixel
    Returns:
        dictionary that maps the HealpixPixel (a pixel at destination order) to a tuple of
        origin pixel information.

        The tuple contains the following:

        - 0 - the total number of rows found in this destination pixel
        - 1 - the set of indexes in histogram for the pixels at the original healpix order.
    Raises:
        ValueError: if the histogram is the wrong size, or some initial histogram bins
            exceed threshold.
    """
    if len(histogram) != hp.order2npix(highest_order):
        raise ValueError("histogram is not the right size")
    if lowest_order > highest_order:
        raise ValueError("lowest_order should be less than highest_order")
    max_bin = np.amax(histogram)
    if max_bin > threshold:
        raise ValueError(f"single pixel count {max_bin} exceeds threshold {threshold}")

    nested_sums = []

    for order in range(0, highest_order):
        explosion_factor = 4 ** (highest_order - order)
        nested_sums.append(histogram.reshape(-1, explosion_factor).sum(axis=1))
    nested_sums.append(histogram)

    ## Determine the lowest order that does not exceed the threshold
    orders_at_threshold = [lowest_order if 0 < k <= threshold else None for k in nested_sums[lowest_order]]
    for order in range(lowest_order + 1, highest_order + 1):
        new_orders_at_threshold = np.array(
            [order if 0 < k <= threshold else None for k in nested_sums[order]]
        )
        parent_alignment = np.repeat(orders_at_threshold, 4, axis=0)
        orders_at_threshold = [
            parent_order if parent_order is not None else new_order
            for parent_order, new_order in zip(parent_alignment, new_orders_at_threshold)
        ]

    ## Zip up the orders and the pixel numbers.
    healpix_pixels = np.array(
        [
            HealpixPixel(order, pixel >> 2 * (highest_order - order)) if order is not None else None
            for order, pixel in zip(orders_at_threshold, np.arange(0, len(orders_at_threshold)))
        ]
    )

    ## Create a map from healpix pixel to origin pixel data
    non_none_elements = healpix_pixels[healpix_pixels != np.array(None)]
    unique_pixels = np.unique(non_none_elements)

    result = {}
    for healpix_pixel in unique_pixels:
        # Find all nonzero constituent pixels at the histogram's order
        explosion_factor = 4 ** (highest_order - healpix_pixel.order)
        start_pixel = healpix_pixel.pixel * explosion_factor
        end_pixel = (healpix_pixel.pixel + 1) * explosion_factor

        non_zero_indexes = np.nonzero(histogram[start_pixel:end_pixel])[0] + start_pixel
        result[healpix_pixel] = (
            nested_sums[healpix_pixel.order][healpix_pixel.pixel],
            non_zero_indexes.tolist(),
        )
    return result


def generate_constant_pixel_map(histogram, constant_healpix_order):
    """Special case of creating a destination pixel map where you want to
    preserve some constant healpix order across the sky.

    NB:
        - this method filters out pixels with no counts
        - no upper thresholds are applied, and single pixel may contain many rows

    Args:
        histogram (:obj:`np.array`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        constant_healpix_order (int):  the desired healpix order (e.g. 0-10)
    Returns:
        dictionary that maps non-empty bins as HealpixPixel to a tuple of origin pixel
        information.

        The tuple contains the following:

        - 0 - the total number of rows found in this destination pixel, same as the origin bin
        - 1 - the set of indexes in histogram for the pixels at the original healpix
          order, which will be a list containing only the origin pixel.
    Raises:
        ValueError: if the histogram is the wrong size
    """
    if len(histogram) != hp.order2npix(constant_healpix_order):
        raise ValueError("histogram is not the right size")

    non_zero_indexes = np.nonzero(histogram)[0]
    healpix_pixels = [HealpixPixel(constant_healpix_order, pixel) for pixel in non_zero_indexes]

    value_list = [(histogram[pixel], [pixel]) for pixel in non_zero_indexes]

    return dict(zip(healpix_pixels, value_list))
